/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERN.hxx"

const char CAVERNdb_server_c::PUT_MESG = 'p';
const char CAVERNdb_server_c::FETCH_MESG = 'f';
const char CAVERNdb_server_c::COMMIT_MESG = 'c';
const int CAVERNdb_server_c::DEFAULT_TCP_REFLECTOR_PORT = 9500;
const int CAVERNdb_server_c::DEFAULT_UDP_REFLECTOR_PORT = 9550;
const int CAVERNdb_server_c::DEFAULT_MAX_CLIENTS = 64;
const int CAVERNdb_server_c::OK = 1;
const int CAVERNdb_server_c::FAILED = 0;
const int CAVERNdb_server_c::MEM_ALLOC_ERR = 3;
const int CAVERNdb_server_c::TOO_MANY_CLIENTS = 4;

const int CAVERNdb_client_c::FAILED = 0;
const int CAVERNdb_client_c::OK = 1;
const int CAVERNdb_client_c::MEM_ALLOC_ERR = 2;
const int CAVERNdb_client_c::CONNECTION_TERMINATED = 3;
const int CAVERNdb_client_c::IGNORED = 4;
const int CAVERNdb_client_c::UNRELIABLE_CONNECTION = 5;
const int CAVERNdb_client_c::RELIABLE_CONNECTION = 6;
const int CAVERNdb_client_c::MAX_UDP_PACKET_SIZE = 1000;
/*
This is the replacement for the CAVERNsoft key database that was a central part
of the earlier generation of CAVERNsoft. In G2 this is no longer the case.
Here it is significantly simplified with fewer capabilities.

One of the simplifications is that it follows only a client server model and
all connections are reliable TCP connections.

Clients instantiate a CAVERNdb_client_c object to communicate with the
server.  The server is built from 2 main components the keyTool
database that was part of the original CAVERNsoft and the new
CAVERNnet_tcpReflector_c. The client is also built from the keyTool
database however it uses the CAVERNnet_tcpReflectorClient_c to speak
to the server.

Basically the server sits around and waits for incoming connections. The
server also listens for incoming messages from these connections. Some messages it
will reflect to all other connected clients. Some messages it will absorb to do
some processing.

Currently the messages that the server recognizes are:

PUT, FETCH, COMMIT

PUT
---

This message will tell the server to store a piece of data in its database
given the pathname, keyname and the data. When a server receives this message
it first stores it in its database and then forwards this same message to
all other connected clients so that the same database update may occur at
each respective client.

FORMAT:
p+pathsize+pathname+keysize+keyname+datasize+dataname
pathsize=int
pathname=char* array of chars
keysize=int
keyname=char*
datasize=int
dataname=char*

FETCH
-----

This message will tell the server to send the data stored in the server's
database at a specific pathname and keyname, back to the client.
This is an asynchronous operation meaning once the client has issued the fetch
command, it will continue doing its own processing.

Clients also recognize messages. Currently the only one the client recognizes
is the PUT message which follows the same format as the PUT message above.
On receipt of a PUT message the client will store new incoming data into its
own database.

FORMAT:
f+pathsize+pathname+keysize+keyname
pathsize=int
pathname=char*
keysize=int
keyname=char*

COMMIT
------

This tells the server to commit a specific key to the database.

FORMAT:
c+pathsize+pathname+keysize+keyname
pathsize=int
pathname=char*
keysize=int
keyname=char*

--------------------------------------------

Future things to add include:
DONE - if no connection exists or if connection broken keys should continue to work
DONE - autoreconnect
- lock local and remote
- smarter server so that it will deallocate memory that's not needed

*/

CAVERNdb_server_c::~CAVERNdb_server_c()
{
	if (keyToolManager) delete keyToolManager;
	if (server) delete server;
    if (udpreflector) delete udpreflector;
}

int CAVERNdb_server_c :: init(char *defaultDBName, int inPort, int maxCli, int udpReflectorPort)
{
	keyToolManager = new keyToolManager_c(defaultDBName);
	if (keyToolManager == NULL) return CAVERNdb_server_c::MEM_ALLOC_ERR;

	incomingPort = inPort;
	server = new CAVERNnet_tcpReflector_c();
	if (server == NULL) {
		delete keyToolManager;
		keyToolManager = NULL;
		return CAVERNdb_server_c::MEM_ALLOC_ERR;
	}
	int status = server->init(inPort, maxCli);
	if (status != CAVERNnet_tcpReflector_c::OK) {
		return CAVERNdb_server_c::FAILED;
	}
	server->interceptNewConnection(newConnectionHandler, this);
	server->intercept(newDataHandler, this);
    
    udpreflector = new CAVERNnet_udpReflector_c;
    if (udpreflector == NULL) 
        return CAVERNdb_server_c::MEM_ALLOC_ERR;
    udpreflector->setIncomingPort(udpReflectorPort);
    udpreflector->init();
    udpreflector->enableInstrumentation();
    udpreflector->intercept(udpDataHandler, this);
	return CAVERNdb_server_c::OK;
}

int CAVERNdb_server_c :: process()
{
	int status;

	status = server->checkForNewClients();
	if (status == CAVERNnet_tcpReflector_c::TOO_MANY_CLIENTS)
		return CAVERNdb_server_c::TOO_MANY_CLIENTS;
	
	status = server->process();
	if (status != CAVERNnet_tcpReflector_c::OK) return CAVERNdb_server_c::FAILED;
    udpreflector->process();
	return CAVERNdb_server_c::OK;
}

void CAVERNdb_server_c :: newConnectionHandler(CAVERNnet_tcpReflectorClient_c *newClient, void* userData)
{
	//CAVERNdb_server_c *server = (CAVERNdb_server_c*) userData;
}

int CAVERNdb_server_c :: newDataHandler(CAVERNnet_tcpReflectorClient_c *newClient, char** buffer, int *bufferSize, void *userData)
{
	CAVERNdb_server_c *server = (CAVERNdb_server_c*) userData;

	// First byte of buffer is command code.
	char* buf = *buffer;

	CAVERNnet_datapack_c data;
	data.initUnpack(buf, *bufferSize);

	char command;
	data.unpackChar(&command);

	switch(command) {

		// CAVERNdb_server_c::PUT_MESG is for PUT command
	case CAVERNdb_server_c::PUT_MESG:
		server->handlePut(buf, *bufferSize);
		return CAVERNnet_tcpReflector_c::OK;
		// CAVERNdb_server_c::FETCH_MESG is for FETCH command
	case CAVERNdb_server_c::FETCH_MESG:
		server->handleFetch(newClient, buf, *bufferSize);

		// For a fetch the reply is a put sent only to the originating
		// client so the message is not forwarded to anyone else.
		return CAVERNnet_tcpReflector_c::SKIP_DISTRIBUTION;
	case CAVERNdb_server_c::COMMIT_MESG:
		server->handleCommit(newClient, buf, *bufferSize);

		// For a fetch the reply is a put sent only to the originating
		// client so the message is not forwarded to anyone else.
		return CAVERNnet_tcpReflector_c::SKIP_DISTRIBUTION;
	default:
		printf("CAVERNdb_server_c: Unknown incoming command.\n");
		return CAVERNnet_tcpReflector_c::SKIP_DISTRIBUTION;
	}
}

int CAVERNdb_server_c :: udpDataHandler(CAVERNnet_udp_c *newClient, char** buffer, int *bufferSize, void *userData)
{
    CAVERNdb_server_c *server = (CAVERNdb_server_c*) userData;

    // First byte of buffer is command code.
    char* buf = *buffer;

    CAVERNnet_datapack_c data;
    data.initUnpack(buf, *bufferSize);

    char command;
    data.unpackChar(&command);

    switch(command) {
    
    // CAVERNdb_server_c::PUT_MESG is for PUT command
    case CAVERNdb_server_c::PUT_MESG:
        server->handlePut(buf, *bufferSize);
        return CAVERNnet_udpReflector_c::OK;
    
    // The db server right now only accepts the put command via udp, the fetch and commit are done using reliable connections 
    
    default:
        printf("CAVERNdb_server_c: Unknown incoming command.\n");
        return CAVERNnet_udpReflector_c::SKIP_DISTRIBUTION;                         }                                                                           }
        
void CAVERNdb_server_c :: handlePut(char *buf, int size)
{

	CAVERNnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname =NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("CAVERNdb_server_c::handlePut: Memory alloc error\n");
			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	// get keyname
	char* keyname = NULL;
	if (keysize) {
		keyname = new char[keysize];
		if (keyname == NULL) {
			printf("CAVERNdb_server_c::handlePut: Memory alloc error\n");
			if (pathsize) delete pathname;
			return;
		}

		data.unpack(keyname, keysize);
	}

	// get datasize
	int datasize;
	data.unpackInt(&datasize);

	char* rawdata = NULL;
	if (datasize) {
		// get data
		rawdata = new char[datasize];

		if (rawdata == NULL) {
			printf("CAVERNdb_server_c::handlePut: Memory alloc error\n");
			if (keysize) delete keyname;
			if (pathsize) delete pathname;
			return;
		}
		data.unpack(rawdata, datasize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	if (aKey == NULL) {
#ifdef WIN32
		printf("CAVERNdb_server_c::handlePut: Cannot access key %s/%s\n",pathname, keyname);
#else
		printf("CAVERNdb_server_c::handlePut: Cannot access key %s\\%s\n",pathname, keyname);
#endif
		if (rawdata) delete rawdata;
		if (keysize) delete keyname;
		if (pathsize) delete pathname;
		return;
	}

	aKey->alloc(datasize);
	char *rawBuffer = aKey->getData();

	if (datasize)
		memcpy(rawBuffer, rawdata, datasize);

	if (datasize)
		delete rawdata;
	delete aKey;

	if (keysize)
		delete keyname;
	if (pathsize)
		delete pathname;

}

void CAVERNdb_server_c :: handleFetch(CAVERNnet_tcpReflectorClient_c *client, char *buf, int size)
{

	CAVERNnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname = NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("CAVERNdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
	// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize) delete pathname;
			printf("CAVERNdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(keyname, keysize);
	}

	sendPutMessage(client, pathname, pathsize, keyname, keysize);

	if (keysize)
		delete keyname;
	if (pathsize)
		delete pathname;
}

void CAVERNdb_server_c :: handleCommit(CAVERNnet_tcpReflectorClient_c *client, char *buf, int size)
{

	CAVERNnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname = NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("CAVERNdb_server_c::handleCommit: Memory alloc error\n");

			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
	// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize) delete pathname;
			printf("CAVERNdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(keyname, keysize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		return;
	}

	aKey->commit();

	if (keysize)
		delete keyname;
	if (pathsize)
		delete pathname;
}

void CAVERNdb_server_c :: sendPutMessage(CAVERNnet_tcpReflectorClient_c *client, char* pathname, int pathsize, char* keyname, int keysize)
{
	// Generate a reply message to send back.

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	if (aKey == NULL) {
#ifdef WIN32
		printf("CAVERNdb_server_c::sendPutMessage: Cannot access key %s/%s\n",pathname, keyname);
#else
		printf("CAVERNdb_server_c::sendPutMessage: Cannot access key %s\\%s\n",pathname, keyname);
#endif
		return;
	}


	char* thedata = aKey->getData();
	int thedatasize = aKey->getDataSize();

	int replyBufferSize = CAVERNnet_datapack_c::sizeof_char()+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(pathsize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(keysize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(thedatasize);

	char *replyBuffer = new char[replyBufferSize];

	if (replyBuffer == NULL) {
		delete aKey;
		printf("CAVERNdb_server_c::sendPutMessage: Memory alloc error\n");
		return;
	}

	CAVERNnet_datapack_c reply;

	reply.initPack(replyBuffer,replyBufferSize);
	reply.packChar(CAVERNdb_server_c::PUT_MESG);
	reply.packInt(pathsize);
	if (pathsize)
		reply.pack(pathname, pathsize);
	reply.packInt(keysize);
	if (keysize)
		reply.pack(keyname, keysize);
	reply.packInt(thedatasize);

	if (thedatasize)
		reply.pack(thedata, thedatasize);

	client->write(replyBuffer, &replyBufferSize);
	delete replyBuffer;
	delete aKey;
}

void CAVERNdb_server_c:: showStats(char* streamInfo, char* comment)
{
    if (server) server->showStats(streamInfo, comment);    
}

int CAVERNdb_server_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return (server) ? server->logStats(streamInfo, comment, filePtr) : CAVERNnet_perfMonitor_c::FAILED;
}

int CAVERNdb_server_c::sendStats(char* streamInfo, char* comment)
{
    return (server) ? server->sendStats(streamInfo, comment) : CAVERNnet_perfMonitor_c::FAILED;
}

int CAVERNdb_server_c::initSendStats(char* monitorClientIP, int port)
{
    return (server) ? server->initSendStats(monitorClientIP,port) : CAVERNnet_perfMonitor_c::FAILED;
}

void CAVERNdb_server_c::exitSendStats()
{
    if (server) server->exitSendStats();
}


/*************************************************************************/


CAVERNdb_client_c :: CAVERNdb_client_c()
{
	client = NULL;
    udpClient = NULL;
	triggerCallback = NULL;
	callbackUserData = NULL;
}

CAVERNdb_client_c :: ~CAVERNdb_client_c()
{
	delete clientmutex;
	delete callbackmutex;
	if (client){
		client->close(); 
		delete client;
	}
    if (udpClient)
        delete udpClient;
	if (keyToolManager) {
		delete keyToolManager;
	}
}

int CAVERNdb_client_c :: init(char* servername, int serverport, char *dbname, int udp_reflector_port)
{

	clientmutex = new CAVERNts_mutex_c;
	callbackmutex = new CAVERNts_mutex_c;

	if (clientmutex == NULL) {
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}
	if (callbackmutex == NULL) {
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}


	keyToolManager = new keyToolManager_c(dbname);

	if (keyToolManager == NULL) {
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	itsServerPort = serverport;
    udpReflectorPort = udp_reflector_port;
	itsServerName[0]='\0';

	if (servername == NULL)
		return OK;

	strcpy(itsServerName, servername);
    udpClient = new CAVERNnet_udp_c;
    if (udpClient == NULL)
        return CAVERNdb_client_c::MEM_ALLOC_ERR;
	udpClient->init();
    udpClient->setSendAddress(itsServerName, udpReflectorPort);
    udpClient->enableInstrumentation();
    udpClient->makeNonBlocking();

    return reconnect();

	// Worry about autoreconnect later
}

int CAVERNdb_client_c :: reconnect(char* servername, int serverport)
{
	clientmutex->lock();
	char theserver[256];
	int theport;

	if (servername == NULL)
		strcpy(theserver, itsServerName);
	if (servername) {
		strcpy(theserver, servername);
		strcpy(itsServerName , servername);
	}
	if (serverport == 0)
		theport = itsServerPort;
	if (serverport) {
		theport =serverport;
		itsServerPort = theport;
	}

	if (theserver[0] == '\0') {
		clientmutex->unlock();
		return CAVERNdb_client_c :: IGNORED;
	}

	client = new CAVERNnet_tcpReflectorClient_c;
	if (client == NULL) {
#ifdef WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
		clientmutex->unlock();
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	int status = client->connectToServer(theserver, theport);
	if (status == CAVERNnet_tcpReflectorClient_c :: FAILED){
		delete client;
		client = NULL;
#ifdef WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
		clientmutex->unlock();
		return CAVERNdb_client_c::FAILED;
	}
	printf("CAVERN DB client connected to: %s:%d\n",theserver,theport);

	clientmutex->unlock();
	return CAVERNdb_client_c::OK;

}

int CAVERNdb_client_c :: fetch(char* pathname, char* keyname)
{
	clientmutex->lock();
	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("CAVERNdb_client_c::fetch: Cannot fetch a NULL keyname\n");
		clientmutex->unlock();
		return CAVERNdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	int sendBufferSize = CAVERNnet_datapack_c::sizeof_char() +
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(pathsize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(keysize);

	char *sendBuffer = new char[sendBufferSize];

	CAVERNnet_datapack_c send;
	send.initPack(sendBuffer,sendBufferSize);
	send.packChar(CAVERNdb_server_c::FETCH_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);
	if (keysize)
		send.pack(keyname, keysize);

	int statusc = CAVERNnet_tcpReflectorClient_c::FAILED;
	if (client) 
	statusc = client->write(sendBuffer, &sendBufferSize);

	delete sendBuffer;
	delete aKey;
	clientmutex->unlock();

	if (statusc == CAVERNnet_tcpReflectorClient_c::OK) {
		return CAVERNdb_client_c::OK;
	}
	return CAVERNdb_client_c::FAILED;
}

char* CAVERNdb_client_c :: get(char* pathname, char* keyname, int* datasize)
{

	if (keyname == NULL) {
		printf("CAVERNdb_client_c::get: Cannot get a NULL keyname\n");
		return NULL;
	}

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	*datasize = aKey->getDataSize();
	if (*datasize == 0) {
		clientmutex->unlock();
		return NULL;
	}

	char* data = new char[*datasize];
	if (data == NULL) {
		clientmutex->unlock();
		return NULL;
	}

	char* keydata = aKey->getData();
	memcpy(data, keydata, *datasize);

	clientmutex->unlock();

	return data;
}

int CAVERNdb_client_c :: commit(char* pathname, char* keyname)
{
	if (keyname == NULL) {
		printf("CAVERNdb_client_c::commit: Cannot commit to a NULL keyname\n");
		return CAVERNdb_client_c::FAILED;
	}

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();

		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	aKey->commit();
	clientmutex->unlock();

	return CAVERNdb_client_c::OK;

}

int CAVERNdb_client_c :: remoteCommit(char* pathname, char* keyname)
{
	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("CAVERNdb_client_c::remoteCommit: Cannot remoteCommit a NULL keyname\n");
		clientmutex->unlock();

		return CAVERNdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	int sendBufferSize = CAVERNnet_datapack_c::sizeof_char() +
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(pathsize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(keysize);

	char *sendBuffer = new char[sendBufferSize];

	CAVERNnet_datapack_c send;
	send.initPack(sendBuffer,sendBufferSize);

	send.packChar(CAVERNdb_server_c::COMMIT_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);
	if (keysize)
		send.pack(keyname, keysize);

	int statusc = CAVERNnet_tcpReflectorClient_c::FAILED;
	if (client) 
		statusc = client->write(sendBuffer, &sendBufferSize);

	delete sendBuffer;
	delete aKey;

	clientmutex->unlock();

	if (statusc == CAVERNnet_tcpReflectorClient_c::OK) {
		return CAVERNdb_client_c::OK;
	}
	return CAVERNdb_client_c::FAILED;
}

int CAVERNdb_client_c :: put(char* pathname, char* keyname, char* data, int thedatasize, int typeOfConnection)
{
	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("CAVERNdb_client_c::put: Cannot put to a NULL keyname\n");
		return CAVERNdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();

		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	aKey->alloc(thedatasize);

	char* thedata = aKey->getData();

	if (thedatasize)
		memcpy(thedata, data, thedatasize);

	int sendBufferSize = CAVERNnet_datapack_c::sizeof_char() +
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(pathsize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(keysize)+
		CAVERNnet_datapack_c::sizeof_int()+
		CAVERNnet_datapack_c::sizeof_char(thedatasize);

	char *sendBuffer = new char[sendBufferSize];
	if (sendBuffer == NULL) {
		delete aKey;
		clientmutex->unlock();
		return CAVERNdb_client_c::MEM_ALLOC_ERR;
	}

	CAVERNnet_datapack_c send;

	send.initPack(sendBuffer,sendBufferSize);
	send.packChar(CAVERNdb_server_c::PUT_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);

	if (keysize)
		send.pack(keyname, keysize);
	send.packInt(thedatasize);

	if (thedatasize)
		send.pack(thedata, thedatasize);
        
    if (typeOfConnection == CAVERNdb_client_c::RELIABLE_CONNECTION)
    {
	    if (client) 
		    client->write(sendBuffer, &sendBufferSize);
    }
    else            //Unreliable udp is to be used
    {
        if (thedatasize <= CAVERNdb_client_c::MAX_UDP_PACKET_SIZE)
        {
            if (udpClient)
                udpClient->send(sendBuffer, sendBufferSize); 
        }
        else
        {
            printf("\nPacket size exceeds 1 KB, by-passing UDP connection and sending it via TCP\n");
            if (client)
               client->write(sendBuffer, &sendBufferSize);
        }
    }
	clientmutex->unlock();

	delete sendBuffer;
	delete aKey;

	return CAVERNdb_client_c::OK;
}

int CAVERNdb_client_c :: process()
{
	int status;
	char* data, *udpData;
	int dataSize;
    	int udpDataSize = CAVERNdb_client_c::MAX_UDP_PACKET_SIZE;
    	int udpReturnSize;
    	udpData = new char[udpDataSize+1];
	clientmutex->lock();

    	if (udpClient)
	{
    		udpData = new char[udpDataSize+1];
        	udpReturnSize = udpClient->receive(udpData, udpDataSize);
        	if (udpReturnSize > 0)
            		handleMessage(udpData, udpDataSize);
        	delete udpData;
	}
	if (client)
		status = client->read(&data, &dataSize, CAVERNnet_tcpReflectorClient_c::NON_BLOCKING);
	clientmutex->unlock();

    	if (udpReturnSize > 0)
        	handleMessage(udpData, udpDataSize);
	if (client == NULL) return CAVERNdb_client_c::CONNECTION_TERMINATED;

	if (status == CAVERNnet_tcpReflectorClient_c :: OK) {
		handleMessage(data, dataSize);
		delete data;
		return CAVERNdb_client_c::OK;
	}

	clientmutex->lock();
	if (status == CAVERNnet_tcpReflectorClient_c :: CONNECTION_TERMINATED) {
		delete client;
		client = NULL;
		clientmutex->unlock();

		return CAVERNdb_client_c::CONNECTION_TERMINATED;
	}

	clientmutex->unlock();

	return CAVERNdb_client_c::OK;

}

void CAVERNdb_client_c :: handleMessage(char* data, int dataSize)
{
	CAVERNnet_datapack_c thepack;
	thepack.initUnpack(data,dataSize);

	char command;
	thepack.unpackChar(&command);

	switch(command) {
	case CAVERNdb_server_c::PUT_MESG:
		handleIncomingPut(data, dataSize);
		break;
	default:
		printf("CAVERNdb_client_c: Unknown incoming command.\n");
	}
}

void CAVERNdb_client_c :: handleIncomingPut(char* buf, int bufsize)
{
	clientmutex->lock();

	CAVERNnet_datapack_c data;

	data.initUnpack(buf,bufsize);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	char* pathname =NULL;

	if (pathsize) {
	// get pathname
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("CAVERNdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}
		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
		// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize) delete pathname;
			printf("CAVERNdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}

		data.unpack(keyname, keysize);
	}

	// get datasize
	int datasize;
	data.unpackInt(&datasize);

	char* rawdata = NULL;

	if (datasize) {
		// get data
		rawdata = new char[datasize];
		if (keyname == NULL) {
			if (pathsize) delete pathname;
			if (keysize) delete keyname;
			printf("CAVERNdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}

		data.unpack(rawdata, datasize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		if (pathsize) delete pathname;
		if (keysize) delete keyname;
		if (datasize) delete rawdata;
#ifdef WIN32
		printf("CAVERNdb_client_c::handleIncomingPut:  Cannot access key %s\\%s\n",pathname, keyname);
#else
		printf("CAVERNdb_client_c::handleIncomingPut:  Cannot access key %s/%s\n",pathname, keyname);
#endif
		clientmutex->unlock();

		return;
	}

	aKey->alloc(datasize);
	char *rawBuffer = aKey->getData();

	if (datasize) {
		memcpy(rawBuffer, rawdata, datasize);
		delete rawdata;
	}

	delete aKey;


	notifyPathname = pathname;
	notifyKeyname = keyname;

	clientmutex->unlock();

	notify();

	if (triggerCallback) {
		callbackmutex->lock();
		triggerCallback(pathname, keyname, callbackUserData);
		callbackmutex->unlock();
	}
	if (keysize) 
		delete keyname;

	if (pathsize)
		delete pathname;


}

void CAVERNdb_client_c :: trigger(void (*callback) (char* pathname, char* keyname, void *userData), void* userData)
{
	clientmutex->lock();
	triggerCallback = callback;
	callbackUserData = userData;
	clientmutex->unlock();

}

void CAVERNdb_client_c::showStats(char* streamInfo, char* comment)
{
    if (client) client->showStats(streamInfo, comment);    
}

int CAVERNdb_client_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return (client) ? client->logStats(streamInfo, comment, filePtr) : CAVERNnet_perfMonitor_c::FAILED;
}

int CAVERNdb_client_c::sendStats(char* streamInfo, char* comment)
{
    return (client) ? client->sendStats(streamInfo, comment) : CAVERNnet_perfMonitor_c::FAILED;
}

int CAVERNdb_client_c::initSendStats(char* monitorClientIP, int port)
{
    return (client) ? client->initSendStats(monitorClientIP,port) : CAVERNnet_perfMonitor_c::FAILED;
}

void CAVERNdb_client_c::exitSendStats()
{
    if (client) client->exitSendStats();
}
