/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERN.hxx"
#include <stdlib.h>

#ifdef WIN32
#include <fcntl.h>
#endif

#ifdef CAVERN_USE_PTHREADS
#include <pthread.h>
#endif

#define CAVERN_VERSION "v1.2.2"

#ifdef sproc_VERSION
#define CAVERN_THREAD_VERSION "SPROC VERSION WITH GLOBUS"
#else
#ifdef CAVERN_USE_WINDOWSTHREADS
#define CAVERN_THREAD_VERSION "MICROSOFT WINDOWS THREAD VERSION WITHOUT GLOBUS"
#elif defined(CAVERN_DO_NOT_USE_GLOBUS)
#define CAVERN_THREAD_VERSION "PTHREAD VERSION WITHOUT GLOBUS"
#else
#define CAVERN_THREAD_VERSION "PTHREAD VERSION WITH GLOBUS"
#endif
#endif

#define CAVERN_VERSION_DATE "Jun 08 2001"
#define _CVRN_MAX_SPROCS 64
#define _CVRN_MAX_SPROC_MUTEX_MEM (1024*1024*4)  // Four Megs default

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#include <sys\timeb.h>

// This will be used in CAVERNmisc_debug.cxx file for internal use only.
CRITICAL_SECTION Win32StdioLock;
LARGE_INTEGER Frequency;
LARGE_INTEGER StartCounter;
struct timeval StartTimeVal;
int UseFTimeOnly = 0;
#endif

#if defined(CAVERN_DO_NOT_USE_GLOBUS) && defined(CAVERN_USE_PTHREADS)
pthread_mutex_t pthread_stdio_lock;
#endif

// Show the version # of CAVERNsoft.
static void version()
{
	//	char date[256], dummy1[256], dummy2[256];
	//	char versionNum[256];
	//	sscanf(rcsid,"%s %s %s %s", dummy1, dummy2, versionNum, date);
	printf("CAVERNsoft G2 %s (%s) [%s]\n\n",  CAVERN_VERSION,
		 CAVERN_VERSION_DATE, CAVERN_THREAD_VERSION);

}

// Check the CAVERN_DEBUG environment variable for debug level to use.
void checkDebugLevel()
{
	// Figure out how verbose CAVERN will be. 0 = user level messages.
	char* debugLevel;
	debugLevel = getenv("CAVERN_DEBUG");

	if (debugLevel)
		cvrnSetDebugFilter(atoi(debugLevel));
	else
		cvrnSetDebugFilter(0);
}

#ifndef CAVERN_DO_NOT_USE_GLOBUS
void cvrnSetGlobusMaxMutexMem(unsigned long maxSprocMutex)
{
	char tmpStr[1024];
	sprintf(tmpStr,"GLOBUS_THREAD_SPROC_MAX_US_MEMORY=%u",maxSprocMutex);
	putenv(tmpStr);
#ifdef sproc_VERSION
	printf("CAVERN: %s\n",tmpStr);
#endif
}

void cvrnSetGlobusMaxThread(int maxSproc)
{
	char tmpStr[1024];
	sprintf(tmpStr,"GLOBUS_THREAD_SPROC_MAX_SPROCS=%d",maxSproc);
	putenv(tmpStr);

#ifdef sproc_VERSION
	printf("CAVERN: %s\n",tmpStr);
#endif

}
#endif

#ifdef WIN32
static int WinSockInitialized = 0;

// This function is going to be registered using atexit function so that
// it is going to be called whenever the program exits.
static void ExitApplication()
{
    if (WinSockInitialized) WSACleanup();
	DeleteCriticalSection(&Win32StdioLock);
}

static void InitWinsock()
{
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
 
    wVersionRequested = MAKEWORD(1, 1); // It requires Winsock version 2.
 
    err = WSAStartup(wVersionRequested, &wsaData);
    if (err != 0) {
		printf("Cannot load the valid winsock 2 dll. Program will be terminated.\n");
        exit(1);
    }
    WinSockInitialized = 1; // Winsock dll is loaded.

	// This will register ExitApplication function to exit chain. By 
	// registering this function, it will be called when the program 
	// exits normally, i.e. not crashing in the middle of execution.
    atexit(ExitApplication);
}
#endif

// Initialize CAVERN.
void CAVERNinit()
{
#ifdef WIN32
    InitializeCriticalSection(&Win32StdioLock);
    InitWinsock();
    _fmode = _O_BINARY; // set reading file mode to binary by default.

    if (!QueryPerformanceFrequency(&Frequency)) {
        printf("High performance counter is not available. ftime function will be used instead to calculate performance.\n");
        UseFTimeOnly = 1;
    }
    else {
        struct _timeb timeSec;

        QueryPerformanceCounter(&StartCounter);
        _ftime(&timeSec);
        StartTimeVal.tv_sec = timeSec.time;
        StartTimeVal.tv_usec = (long) timeSec.millitm;
        StartTimeVal.tv_usec *= 1000;
    }        
#endif

	version();
	checkDebugLevel();

#ifndef CAVERN_DO_NOT_USE_GLOBUS
	char *envMaxSproc = getenv("CAVERN_MAX_THREAD");
	int maxSproc;
	if (envMaxSproc) maxSproc = atoi(envMaxSproc);
	else
		maxSproc = _CVRN_MAX_SPROCS;
	cvrnSetGlobusMaxThread(maxSproc);

	char *envMaxSprocMutex = getenv("CAVERN_MAX_MUTEX_MEM");
	unsigned long maxSprocMutex;
	if (envMaxSprocMutex) maxSprocMutex = atoi(envMaxSprocMutex);
	else
		maxSprocMutex = _CVRN_MAX_SPROC_MUTEX_MEM;
	cvrnSetGlobusMaxMutexMem(maxSprocMutex);

	if (globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
		fprintf(stderr,"Globus Common Module activation failed.\n");
		fprintf(stderr,"CAVERN cannot launch.\n");
		exit(1);
	}
#endif

#if defined(CAVERN_DO_NOT_USE_GLOBUS) && defined(CAVERN_USE_PTHREADS)
    pthread_mutex_init(&pthread_stdio_lock, NULL);
#endif

	return;
}

void CAVERNexit()
{
#ifndef CAVERN_DO_NOT_USE_GLOBUS
	if (globus_module_deactivate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
		fprintf(stderr,"Globus Common Module deactivation failed.\n");
		exit(1);
	}
#endif
#if defined(CAVERN_DO_NOT_USE_GLOBUS) && defined(CAVERN_USE_PTHREADS)
    pthread_mutex_destroy(&pthread_stdio_lock);
#endif
}
