/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include <string.h>
#include "CAVERNnet_datapack_c.hxx"

const int CAVERNnet_datapack_c::FAILED = -1;
const int CAVERNnet_datapack_c::OK = 0;

/* 1. CAVERN_LITTLE_ENDIAN_FORMAT must be defined for all 
      machines that uses little endian byte ordering scheme.
      Most popular machines using little endian are Intel and
      Alpha processor. 

   2. Note that using different byte order is not dependent on
      Operating system. It is closely tied to the processor type
	  the system is based on. Thus, linux which runs on different
	  processors may have different byte ordering scheme.

   3. With a simple test program, it looks like that at least
      Sparc machines from SUN, SGI, and Intel use the same 
	  format to store the floating point values except they must
	  be stored using different byte ordering scheme. Probably
	  they are all following the IEEE standard to store all 
	  floating point values.

   4. Thus, if you want to port these functions to some other
      platforms with different processors, you have to carefully
	  check what format they are using to store floating point 
	  values and how each byte is stored in memory for the specific
	  processors.
*/

typedef union {
	int value;
	unsigned char bytes[INT_SIZE];
} INT_VALUE;

typedef union {
	int32 value;
	unsigned char bytes[INT32_SIZE];
} INT32_VALUE;

typedef union {
	int64 value;
	unsigned char bytes[INT64_SIZE];
} INT64_VALUE;

typedef union {
	long value;
	unsigned char bytes[LONG_SIZE];
} LONG_VALUE;

typedef union {
	float value;
	unsigned char bytes[FLOAT_SIZE];
} FLOAT_VALUE;

typedef union {
	double value;
	unsigned char bytes[DOUBLE_SIZE];
} DOUBLE_VALUE;

void CAVERNnet_datapack_c::initPack(char *buffer, int buffersize){
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	nexus_byte_t machine_format;

	start = (nexus_byte_t *)buffer;
	running = (nexus_byte_t *)buffer;
	size = buffersize;

	machine_format=NEXUS_DC_FORMAT_LOCAL;
	nexus_dc_put_byte(&running,&machine_format,1);

	//This line put here so that we don't unpack by accident
	sender_format=-1;
#else
	start = (unsigned char *) buffer;
	running = (unsigned char *) buffer;
	size = buffersize;
#endif
}

void CAVERNnet_datapack_c::initUnpack(char *buffer, int buffersize){
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	nexus_byte_t machine_format;

	start = (nexus_byte_t *)buffer;
	running = (nexus_byte_t *)buffer;
	size = buffersize;

	nexus_dc_get_byte(&running,&machine_format,1,NEXUS_DC_FORMAT_LOCAL);
	sender_format=(int)machine_format;
#else
	start = (unsigned char *) buffer;
	running = (unsigned char *) buffer;
	size = buffersize;
#endif
}


/// Floats 
int CAVERNnet_datapack_c::packFloat(float val)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	float localval=val;
	if (OK==checkspace(  nexus_dc_sizeof_float(1)  )  ){
		nexus_dc_put_float(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else // if USE_GLOBUS_DATA_PACK is undefined
	FLOAT_VALUE oldVal;

	if (checkspace(FLOAT_SIZE) != OK) return FAILED; // not enough buffer space

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	// reverse the order to make the format to BigEndian
	for (i = FLOAT_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, FLOAT_SIZE);
#endif // CAVERN_LITTLE_ENDIAN_FORMAT

	running += FLOAT_SIZE; // increment the pointer
	return OK;
#endif

}

/// Int
int CAVERNnet_datapack_c::packInt(int val) 
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	int localval=val;
	if (OK==checkspace(nexus_dc_sizeof_int(1))){
		nexus_dc_put_int(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT_VALUE oldVal;

	if (checkspace(INT_SIZE) != OK) return FAILED;

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	for (i = INT_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, INT_SIZE);
#endif

	running += INT_SIZE;
	return OK;
#endif
}

/// Int32
int CAVERNnet_datapack_c::packInt32(int32 val) 
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	int localval=val;
	if (OK==checkspace(nexus_dc_sizeof_int(1))){
		nexus_dc_put_int(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT32_VALUE oldVal;

	if (checkspace(INT32_SIZE) != OK) return FAILED;

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	for (i = INT32_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, INT32_SIZE);
#endif

	running += INT32_SIZE;
	return OK;
#endif
}

/// Long
int CAVERNnet_datapack_c::packLong(long val) 
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	long localval=val;
	if (OK==checkspace(nexus_dc_sizeof_long(1))){
		nexus_dc_put_long(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	LONG_VALUE oldVal;

	if (checkspace(LONG_SIZE) != OK) return FAILED;

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	for (i = LONG_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, LONG_SIZE);
#endif

	running += LONG_SIZE;
	return OK;
#endif
}

/// Int64
int CAVERNnet_datapack_c::packInt64(int64 val) 
{
	INT64_VALUE oldVal;

	if (checkspace(INT64_SIZE) != OK) return FAILED;

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	for (i = INT64_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, INT64_SIZE);
#endif

	running += INT64_SIZE;
	return OK;
}

/// Double
int CAVERNnet_datapack_c::packDouble(double val)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	double localval=val;
	if (OK==checkspace(nexus_dc_sizeof_double(1))){
		nexus_dc_put_double(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	DOUBLE_VALUE oldVal;

	if (checkspace(DOUBLE_SIZE) != OK) return FAILED;

	oldVal.value = val;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;

	for (i = DOUBLE_SIZE - 1, j = 0; i >= 0; i--, j++)
		running[j] = oldVal.bytes[i];
#else
	memcpy(running, oldVal.bytes, DOUBLE_SIZE);
#endif

	running += DOUBLE_SIZE;
	return OK;
#endif
}


/// Char
int CAVERNnet_datapack_c::packChar(char val) 
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	char localval=val;
	if (OK==checkspace(nexus_dc_sizeof_char(1))){
		nexus_dc_put_char(&running,&localval,1);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	if (checkspace(CHAR_SIZE) != OK) 
		return FAILED;

	*running++ = (unsigned char) val;
	return OK;
#endif
}

int CAVERNnet_datapack_c::pack(char* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_char(sz))){
		nexus_dc_put_char(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	if (checkspace(CHAR_SIZE * sz) != OK) return FAILED;

	memcpy(running, val, CHAR_SIZE * sz);
	running += (CHAR_SIZE * sz);
	return OK;
#endif
}

//pack an array of floats
int CAVERNnet_datapack_c::packFloatArray(float* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_float(sz))){
		nexus_dc_put_float(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
#endif

	FLOAT_VALUE* oldArray = new FLOAT_VALUE[sz];       
    if (checkspace(FLOAT_SIZE * sz) != OK) return FAILED;
    //For each float, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= FLOAT_SIZE-1, j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, FLOAT_SIZE);
        #endif        
	    running += (FLOAT_SIZE);
        
     }
    delete oldArray; 
    return OK;
#endif
}

//pack an array of doubles
int CAVERNnet_datapack_c::packDoubleArray(double* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_double(sz))){
		nexus_dc_put_double(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
#endif

    DOUBLE_VALUE* oldArray = new DOUBLE_VALUE[sz];       
    if (checkspace(DOUBLE_SIZE * sz) != OK) return FAILED;
    
    //For each double, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= DOUBLE_SIZE-1, j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, DOUBLE_SIZE);
        #endif        
	    running += (DOUBLE_SIZE);
        
     }   
    delete oldArray;
    return OK;
#endif
}

//pack an array of int values
int CAVERNnet_datapack_c::packIntArray(int* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_put_int(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT_VALUE* oldArray = new INT_VALUE[sz];       
    if (checkspace(INT_SIZE * sz) != OK) return FAILED;
   
#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
#endif   
    
    //For each int value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= INT_SIZE-1, j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, INT_SIZE);
        #endif        
	    running += (INT_SIZE);
        
     }   
    delete oldArray;
    return OK;
#endif
}

//pack an array of int32 values
int CAVERNnet_datapack_c::packInt32Array(int32* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_put_int(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT32_VALUE* oldArray = new INT32_VALUE[sz];       
    if (checkspace(INT32_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
#endif
    
    //For each int32 value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= INT32_SIZE-1,j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, INT32_SIZE);
        #endif        
	    running += (INT32_SIZE);
        
     }   
    delete oldArray;
    return OK;
#endif
}

//pack an array of int64 values
int CAVERNnet_datapack_c::packInt64Array(int64* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_put_int(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT64_VALUE* oldArray = new INT64_VALUE[sz];       
    if (checkspace(INT64_SIZE * sz) != OK) return FAILED;
    
#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;   
#endif

    //For each int64 value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= INT64_SIZE-1, j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, INT64_SIZE);
        #endif        
	    running += (INT64_SIZE);
        
     }   
    delete oldArray;
    return OK;
#endif
}

//pack an array of long values
int CAVERNnet_datapack_c::packLongArray(long* val, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_long(sz))){
		nexus_dc_put_long(&running,val,sz);
		return OK;
	}
	else{
		return FAILED;
	}
#else
	LONG_VALUE* oldArray = new LONG_VALUE[sz];       
    if (checkspace(LONG_SIZE * sz) != OK) return FAILED;
    
#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;   
#endif
     
    //For each long value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {
        oldArray[count].value = val[count];
        
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            for(i= LONG_SIZE-1, j=0; i>=0; i--, j++)
                running[j] = oldArray[count].bytes[i];
        #else
            memcpy(running, oldArray[count].bytes, LONG_SIZE);
        #endif        
	    running += (LONG_SIZE);
        
     }   
    delete oldArray;
    return OK;
#endif
}

/// Float
int CAVERNnet_datapack_c::unpackFloat(float *Answer) 
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_float(1))){
		nexus_dc_get_float(&running,Answer,1,sender_format);
		 return OK;
	}else{
		return FAILED;
	}
#else
	FLOAT_VALUE newVal;

	if (checkspace(FLOAT_SIZE) != OK) return FAILED; // not enough buffer space

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[FLOAT_SIZE];

	memcpy(buffer, running, FLOAT_SIZE);

	// reverse the order to make the format from BigEndian to LittleEndian
	for (i = FLOAT_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, FLOAT_SIZE);
#endif // CAVERN_LITTLE_ENDIAN_FORMAT

	*Answer = newVal.value;
	running += FLOAT_SIZE; // increment the pointer
	return OK;
#endif
}

/// Int
int CAVERNnet_datapack_c::unpackInt(int *Answer)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(1))){
		nexus_dc_get_int(&running,Answer,1,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	INT_VALUE newVal;

	if (checkspace(INT_SIZE) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[INT_SIZE];

	memcpy(buffer, running, INT_SIZE);

	for (i = INT_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, INT_SIZE);
#endif

	*Answer = newVal.value;
	running += INT_SIZE;
	return OK;
#endif
}

/// Int32
int CAVERNnet_datapack_c::unpackInt32(int32 *Answer)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(1))){
		nexus_dc_get_int(&running,Answer,1,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	INT32_VALUE newVal;

	if (checkspace(INT32_SIZE) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[INT32_SIZE];

	memcpy(buffer, running, INT32_SIZE);

	for (i = INT32_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, INT32_SIZE);
#endif

	*Answer = newVal.value;
	running += INT32_SIZE;
	return OK;
#endif
}

/// Long
int CAVERNnet_datapack_c::unpackLong(long *Answer)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_long(1))){
		nexus_dc_get_long(&running,Answer,1,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	LONG_VALUE newVal;

	if (checkspace(LONG_SIZE) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[LONG_SIZE];

	memcpy(buffer, running, LONG_SIZE);

	for (i = LONG_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, LONG_SIZE);
#endif

	*Answer = newVal.value;
	running += LONG_SIZE;
	return OK;
#endif
}

int CAVERNnet_datapack_c::unpackInt64(int64 *Answer)
{
	INT64_VALUE newVal;

	if (checkspace(INT64_SIZE) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[INT64_SIZE];

	memcpy(buffer, running, INT64_SIZE);

	for (i = INT64_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, INT64_SIZE);
#endif

	*Answer = newVal.value;
	running += INT64_SIZE;
	return OK;
}

/// Double
int CAVERNnet_datapack_c::unpackDouble(double  *Answer)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_double(1))){
		nexus_dc_get_double(&running,Answer,1,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	DOUBLE_VALUE newVal;

	if (checkspace(DOUBLE_SIZE) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
	int i, j;
	char buffer[DOUBLE_SIZE];

	memcpy(buffer, running, DOUBLE_SIZE);

	for (i = DOUBLE_SIZE - 1, j = 0; i >= 0; i--, j++)
		newVal.bytes[j] = buffer[i];
#else
	memcpy(newVal.bytes, running, DOUBLE_SIZE);
#endif

	*Answer = newVal.value;
	running += DOUBLE_SIZE;
	return OK;
#endif
}

/// Char
int CAVERNnet_datapack_c::unpackChar(char *Answer)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_char(1))){
		nexus_dc_get_char(&running,Answer,1,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	if (checkspace(CHAR_SIZE) != OK) return FAILED;

	*Answer = *running++;
	return OK;
#endif
}


int CAVERNnet_datapack_c::unpack(char *Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_char(sz))){
		nexus_dc_get_char(&running,Answer,sz,sender_format);
		return OK;
	}else{
		return FAILED;
	}
#else
	if (checkspace(CHAR_SIZE * sz) != OK) return FAILED;

	memcpy(Answer, running, CHAR_SIZE * sz);
	running += (CHAR_SIZE * sz);
	return OK;
#endif
}

//unpack an array of floats
int CAVERNnet_datapack_c::unpackFloatArray(float* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_float(sz))){
		nexus_dc_get_float(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	FLOAT_VALUE* newArray = new FLOAT_VALUE[sz];       
    if (checkspace(FLOAT_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
    char buffer[FLOAT_SIZE];
#endif

    //For each float, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, FLOAT_SIZE);
            for(i= FLOAT_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, FLOAT_SIZE);
        #endif        
	    running += (FLOAT_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

//unpack an array of doubles
int CAVERNnet_datapack_c::unpackDoubleArray(double* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_double(sz))){
		nexus_dc_get_double(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	DOUBLE_VALUE* newArray = new DOUBLE_VALUE[sz];       
    if (checkspace(DOUBLE_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT    
    int i,j;
    char buffer[DOUBLE_SIZE];
#endif

    //For each value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, DOUBLE_SIZE);
            for(i= DOUBLE_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, DOUBLE_SIZE);
        #endif        
	    running += (DOUBLE_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

//unpack an array of int values
int CAVERNnet_datapack_c::unpackIntArray(int* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_get_int(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT_VALUE* newArray = new INT_VALUE[sz];       
    if (checkspace(INT_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
    char buffer[INT_SIZE];
#endif

    //For each value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, INT_SIZE);
            for(i= INT_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, INT_SIZE);
        #endif        
	    running += (INT_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

//unpack an array of int32 values
int CAVERNnet_datapack_c::unpackInt32Array(int32* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_get_int(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT32_VALUE* newArray = new INT32_VALUE[sz];       
    if (checkspace(INT32_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT
    int i,j;
    char buffer[INT32_SIZE];
#endif
 
    //For each value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, INT32_SIZE);
            for(i= INT32_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, INT32_SIZE);
        #endif        
	    running += (INT32_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

//unpack an array of int64 values
int CAVERNnet_datapack_c::unpackInt64Array(int64* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_int(sz))){
		nexus_dc_get_int(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	INT64_VALUE* newArray = new INT64_VALUE[sz];       
    if (checkspace(INT64_SIZE * sz) != OK) return FAILED;

#ifdef CAVERN_LITTLE_ENDIAN_FORMAT 
    int i,j;
    char buffer[INT64_SIZE];
#endif    
    
    //For each value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, INT64_SIZE);
            for(i= INT64_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, INT64_SIZE);
        #endif        
	    running += (INT64_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

//unpack an array of long values
int CAVERNnet_datapack_c::unpackLongArray(long* Answer, int sz)
{
#ifdef CAVERN_USE_GLOBUS_DATA_PACK
	if (OK==checkspace(nexus_dc_sizeof_long(sz))){
		nexus_dc_get_long(&running,Answer,sz, sender_format)
		return OK;
	}
	else{
		return FAILED;
	}
#else
	LONG_VALUE* newArray = new LONG_VALUE[sz];       
    if (checkspace(LONG_SIZE * sz) != OK) return FAILED;
    
#ifdef CAVERN_LITTLE_ENDIAN_FORMAT    
    int i,j;
    char buffer[LONG_SIZE];
#endif    
    
    //For each long value, copy the bytes and reverse the ordering for little endian conversion
    for (int count=0; count<sz; count++)
    {   
        #ifdef CAVERN_LITTLE_ENDIAN_FORMAT
            memcpy(buffer, running, LONG_SIZE);
            for(i= LONG_SIZE-1, j=0; i>=0; i--, j++)
                newArray[count].bytes[j] = buffer[i];
        #else
            memcpy(newArray[count].bytes, running, LONG_SIZE);
        #endif        
	    running += (LONG_SIZE); 
        Answer[count] = newArray[count].value;
     }   
    delete newArray;
    return OK;
#endif
}

int CAVERNnet_datapack_c::checkspace(unsigned int incoming_size)
{
	unsigned int remaining;

	remaining=start+size-running;

	if (remaining < incoming_size)
		return FAILED;
	else
		return OK;
}
