/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERNnet_mcast_c.hxx"
#include "perfmon.hxx"

const int CAVERNnet_mcast_c::maxTTL = 128;
const int CAVERNnet_mcast_c::OK = 1;
const int CAVERNnet_mcast_c::FAILED = 0;

#ifdef sun
long gethostid(void);
#endif

#ifdef WIN32
#define MAXHOSTNAME   128   /* Maximum length of a DNS hostname */

// This function is excerpted from "Windows Sockets Network Programming"
// by Bob Quinn and Dave Shute published from Addison-Wesley.
static long gethostid(void)
{
	char szLclHost [MAXHOSTNAME];
	LPHOSTENT lpstHostent;
	SOCKADDR_IN stLclAddr;
	SOCKADDR_IN stRmtAddr;
	int nAddrSize = sizeof(SOCKADDR);
	SOCKET hSock;
	int nRet;

	/* Init local address (to zero) */
	stLclAddr.sin_addr.s_addr = INADDR_ANY;

	/* Get the local hostname */
	nRet = gethostname(szLclHost, MAXHOSTNAME); 
	if (nRet != SOCKET_ERROR) {
		/* Resolve hostname for local address */
		lpstHostent = gethostbyname((LPSTR)szLclHost);
		if (lpstHostent)
			stLclAddr.sin_addr.s_addr = *((u_long FAR*) (lpstHostent->h_addr));
	} 

	/* If still not resolved, then try second strategy */
	if (stLclAddr.sin_addr.s_addr == INADDR_ANY) {
		/* Get a UDP socket */
		hSock = socket(AF_INET, SOCK_DGRAM, 0);
		if (hSock != INVALID_SOCKET)  {
			/* Connect to arbitrary port and address (NOT loopback) */
			stRmtAddr.sin_family = AF_INET;
			stRmtAddr.sin_port   = htons(IPPORT_ECHO);
			stRmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
			nRet = connect(hSock, (LPSOCKADDR)&stRmtAddr,
						   sizeof(SOCKADDR));
			if (nRet != SOCKET_ERROR) {
				/* Get local address */
				getsockname(hSock, (LPSOCKADDR)&stLclAddr, 
							(int FAR*)&nAddrSize);
			}
			closesocket(hSock);   /* we're done with the socket */
		}
	}
	return (stLclAddr.sin_addr.s_addr);
}
#endif

CAVERNnet_mcast_c :: CAVERNnet_mcast_c ()
{
	sockfd = 0;
	enableInstr = 0;
}

void CAVERNnet_mcast_c::enableInstrumentation()
{
	enableInstr = 1;
}

void CAVERNnet_mcast_c::disableInstrumentation()
{
	enableInstr = 0;
}

void CAVERNnet_mcast_c::printInfo()
{
	printf("\nCAVERNPLUS_MCAST_C Info:\n");

	char name[256];

	getSelfIP(name);
	printf("Self Address: %s Port: %u\n",name,
	       getSelfPort());

	getSendIP(name);
	printf("Multicast Send Address: %s Port: %u\n",
	       name,
	       getSendPort());

	getReceiveIP(name);
	printf("Receive Address: %s Port: %u\n\n",
	       name,
	       getReceivePort());
}


int CAVERNnet_mcast_c :: init(char *CAVERNnet_mcast_cGroup, int port, int mttl)
{
	/* create what looks like an ordinary UDP socket */
/* socket function in Windows is returns a unsigned value. Thus, checking 
   whether the result is less than 0 cannot give us accurate result of the
   function call. We have to compare the result with INVALID_SOCKET, instead. */
#ifdef WIN32 
	if ((sockfd=socket(AF_INET,SOCK_DGRAM,0)) == INVALID_SOCKET) {
		printf("CAVERNnet_mcast_c::init: Error in Socket opening.\n");
		return CAVERNnet_mcast_c::FAILED;
	}
#else
	if ((sockfd=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
		printf("CAVERNnet_mcast_c::init: Error in Socket opening.\n");
		return CAVERNnet_mcast_c::FAILED;
	}
#endif

	/* set up destination address */
	memset(&receiveAddress,0,sizeof(receiveAddress));
	receiveAddress.sin_family=AF_INET;
	receiveAddress.sin_addr.s_addr=htonl(INADDR_ANY); /* N.B.: differs from sender */
	receiveAddress.sin_port=htons(port);

	memset(&sendAddress,0,sizeof(sendAddress));
	sendAddress.sin_family=AF_INET;
	sendAddress.sin_addr.s_addr=inet_addr(CAVERNnet_mcast_cGroup);
	sendAddress.sin_port=htons(port);
     
     
	/* bind to receive address */
	if (bind(sockfd,(struct sockaddr *) &receiveAddress,sizeof(receiveAddress)) < 0) {
		printf("CAVERNnet_mcast_c::init: Error in bind.\n");
		return CAVERNnet_mcast_c::FAILED;
	}

	char mcloop = 0;
	/* Disable loopback of CAVERNnet_mcast_c packets so that you don't
	 * get a copy of what you sent.
	 */
	if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_LOOP, &mcloop, sizeof(mcloop)) < 0){
		printf("CAVERNnet_mcast_c::init: setsockopt CAVERNnet_mcast_c cannot set loopback.\n");
	}

	ttl = char(mttl);
	char CAVERNnet_mcast_c_ttl = ttl;

	if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL, &CAVERNnet_mcast_c_ttl, 
		sizeof(CAVERNnet_mcast_c_ttl)) < 0)	{
		printf("CAVERNnet_mcast_c::init: setsockopt CAVERNnet_mcast_c cannot set Time To Live.");
	}

	/* use setsockopt() to request that the kernel join a multicast group */
	mreq.imr_multiaddr.s_addr=inet_addr(CAVERNnet_mcast_cGroup);
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
// Winsock requires the 4th parameter as const char*, and Visual C++ shows
// an error that says passing different type of parameter.
#if (defined(WIN32)) || (defined(sun))
	if (setsockopt(sockfd,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char*) &mreq,sizeof(mreq)) < 0) {
		printf("CAVERNnet_mcast_c::init: setsockopt cannot add membership.\n");
		return CAVERNnet_mcast_c::FAILED;
	}
#else
	if (setsockopt(sockfd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) < 0) {
		printf("CAVERNnet_mcast_c::init: setsockopt cannot add membership.\n");
		return CAVERNnet_mcast_c::FAILED;
	}
#endif

	return CAVERNnet_mcast_c::OK;

}

int CAVERNnet_mcast_c :: send(char *message, int length) 
{

	char* newmsg = message;
	int newlength = length;
	if (enableInstr) {
		newmsg = _cavern_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;

		_cavern_prependPerformanceMonitoringHeader(newmsg, message, length);
	}

	int n = sendto(sockfd, newmsg, newlength, 0, (sockaddr *)&sendAddress,
		       sizeof(sendAddress));

	if (n > 0) incrementDataSent(n);
	if (enableInstr) {
		delete newmsg;
		int realsize = n-_cavern_getSizeOfPerformanceMonitoringHeader();
		if (realsize > 0)
			return realsize;
		else
			return 0;
	}
	else return n;

	/*
	int n;

	n = sendto(sockfd,message,size,0,(struct sockaddr *) &sendAddress,sizeof(sendAddress));
	if (n>0) incrementDataSent(n);
	return n;
	*/
}

int CAVERNnet_mcast_c :: receive(char *message, int length)
{
	char *newmsg = message;
	int newlength = length;

	// if instrumentation enabled
	if (enableInstr) {
		newmsg = _cavern_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;
	}


	int clilen = sizeof(receiveAddress);
	int n;

	n = recvfrom(sockfd, newmsg, newlength, 0, (struct sockaddr *) &receiveAddress,
		     & clilen);
	if (n > 0) {
		incrementDataRead(n);

		// if instrumentation enabled
		if (enableInstr) {
			double remTimeStamp, latency;
			int realSize;
			_cavern_unprependPerformanceMonitoringHeader(&remTimeStamp, &latency, newmsg, n, message, &realSize);

			delete newmsg;

			setInstantLatency(latency);
			return realSize;
		}
	} else {
#ifdef WIN32
		if (n == SOCKET_ERROR) errno = WSAGetLastError();
#endif
		if (enableInstr) delete newmsg;
	}
	
	return n;

	/*
	int  addrlen=sizeof(receiveAddress);
	int n;

	n =recvfrom(sockfd,message,bufferSize,0, (struct sockaddr *) &receiveAddress,&addrlen);
	if (n>0) incrementDataRead(n);

	return n;
	*/

}


void CAVERNnet_mcast_c :: makeNonBlocking() {
	if (sockfd == 0) {
		printf("CAVERNnet_mcast_c::MakeNonBlocking: No connection open yet.\n");
	}
#ifdef __sgi
	fcntl(sockfd, F_SETFL, FNDELAY);
#endif
#ifdef linux
	fcntl(sockfd, F_SETFL, O_NONBLOCK);
#endif
#ifdef sun
	fcntl(sockfd, F_SETFL, O_NDELAY);
#endif
#ifdef WIN32
    u_long nonBlockingMode = 1L;

    ioctlsocket(sockfd, FIONBIO, &nonBlockingMode);
#endif
}

unsigned long CAVERNnet_mcast_c :: getReceiveIP() {
	return receiveAddress.sin_addr.s_addr;
}

void CAVERNnet_mcast_c :: getReceiveIP(char *clientIP) {
	ipNumToString(getReceiveIP(), clientIP);
}

unsigned long CAVERNnet_mcast_c :: getSendIP() {
	return sendAddress.sin_addr.s_addr;
}

void CAVERNnet_mcast_c :: getSendIP(char *clientIP) {
	ipNumToString(getSendIP(), clientIP);
}

void CAVERNnet_mcast_c :: getSelfIP(char *ip) {
	ipNumToString(getSelfIP(), ip);
}
unsigned long CAVERNnet_mcast_c :: getSelfIP() {
#ifdef linux
        unsigned short a1, a2;
        unsigned long ip, temp;

        ip = gethostid();
        a1 = (unsigned short) (ip >> 16);
        a2 = (unsigned short) ip;
        temp = (unsigned long) a2;
        temp <<= 16;
        temp |= a1;
        return temp;
#else
	return gethostid();
#endif
}

unsigned short  CAVERNnet_mcast_c :: getSelfPort() {
	struct sockaddr_in name;
	int size = sizeof(name);
	getsockname (sockfd, (sockaddr *)&name, &size);
	return ntohs(name.sin_port);
}


unsigned short CAVERNnet_mcast_c :: getSendPort() {
	return ntohs(sendAddress.sin_port);
}

unsigned short CAVERNnet_mcast_c :: getReceivePort() {
	return ntohs(receiveAddress.sin_port);
}

