/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERN.hxx"
#include <sys/stat.h>

#ifdef WIN32
#include <stdlib.h>
#define MAXPATHLEN	MAX_PATH
#else
#include <sys/param.h>
#include <netinet/in.h>
#endif

#ifdef sun
#define INADDR_NONE 0xffffffff
#endif

/*
Data stream format from client to server:

Upload:
'u' + 
filenamelen[int] + 
filename +
datasize[int]
data

Download:
'd' +
filenamelen[int] +
filename

Check Remote Timestamp:
'c' +
filenamelen[int] +
filename

Data Stream from server to client:

From server to client- in response to download message:
'd' +
(filesize[int], time-stamp[double]) +
data

From server to client in response to check for remote timestamp:
'c' + 
date[6 integers], timestamp[double]

From Server to Client (For all commands)

File Not found:  'n'
Invalid command: 't'

*/

struct ServerData
{
    char* neighborServerName;
	CAVERNnet_extendedTcpClient_c* client;
	CAVERNnet_tcpServer_c* server;
	int store;
	int port;

	~ServerData() { if (client) { client->close(); delete client; } }
};

/* Minimized client version of remote file IO client that can
   only download.
*/
class CAVERNnet_ServerClient32_c {
	public:

	/// Status ok
	static const int OK/* = 1*/;

	/// Status failed
	static const int FAILED/* = 0*/;

	/// Cannot connect to remote server.
	static const int CANNOT_CONNECT/* = 2*/;

	/// Cannot open local file for reading.
	static const int CANNOT_READ_LOCAL_FILE/* = 3*/;

	/// CAnnot write to local file.
	static const int CANNOT_WRITE_LOCAL_FILE/* = 3*/;

	/// Received invalid command from server.
	static const int INVALID_COMMAND/* = 4*/;

	/// Memory allocation error.
	static const int MEM_ALLOC_ERR/* = 5*/;

	/// Remote file error.
	static const int FILE_ERROR/* = 6*/;

	/// Remote file error.
	static const int DOWNLOAD/* = 7*/;

	CAVERNnet_ServerClient32_c(CAVERNnet_tcpServer_c* Server,
						       int ServerIP = 0, int ServerPort = 0);
	~CAVERNnet_ServerClient32_c();

	/// Set port number to connect to.
	void setRemotePort(int port = CAVERNnet_remoteFileIOServer32_c::DEFAULT_PORT) {
	    neighborPortNumber = port;
	}

	/** Download a file from the remote server. If the remote file is not found or
	    a network error occurs, this function will return with a failed.
	    @param ipAddress is IP address of remote server.
	    @param localFileName is local file write to.
	    @param remoteFileName is remote file to get.
	    @return OK, FAILED, CANNOT_WRITE_LOCAL_FILE, INVALID_COMMAND, CANNOT_CONNECT
	*/
	int initDownload(char *ipAddress, char* localFilename, char* remoteFileName);
	int readBlock(char*, int nBytes);
    int getFileSize(int& fileSize) { return client->readInt(fileSize, CAVERNnet_tcpClient_c::BLOCKING); }

	int getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, 
		        int *remMon, int *remYear, int* remHour, int* remMin, int* remSec);

private:
	int neighborPortNumber;
	int store;
	int serverIPNumber;
	int serverPortNumber;
	char neighborServerName[MAXPATHLEN];
	CAVERNnet_tcpServer_c* server;
	CAVERNnet_extendedTcpClient_c* client;
};

const int CAVERNnet_ServerClient32_c::OK = 1;
const int CAVERNnet_ServerClient32_c::FAILED = 0;
const int CAVERNnet_ServerClient32_c::CANNOT_CONNECT = 2;
const int CAVERNnet_ServerClient32_c::CANNOT_READ_LOCAL_FILE = 3;
const int CAVERNnet_ServerClient32_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int CAVERNnet_ServerClient32_c::INVALID_COMMAND = 4;
const int CAVERNnet_ServerClient32_c::MEM_ALLOC_ERR = 5;
const int CAVERNnet_ServerClient32_c::FILE_ERROR = 6;
const int CAVERNnet_ServerClient32_c::DOWNLOAD = 7;

CAVERNnet_ServerClient32_c::CAVERNnet_ServerClient32_c(CAVERNnet_tcpServer_c* Server,
													   int ServerIP, int ServerPort)
{
	client = NULL;
	server = Server;
	serverIPNumber = ServerIP;
	serverPortNumber = ServerPort;
	neighborPortNumber = CAVERNnet_remoteFileIOServer32_c::DEFAULT_PORT;
}

CAVERNnet_ServerClient32_c::~CAVERNnet_ServerClient32_c()
{
	if (client) {
		client->close();
		delete client;
	}
}

int CAVERNnet_ServerClient32_c::getRemoteDateTime(char *ipAddress, 
						char* remoteFileName, int *remDay, int *remMon, 
						int *remYear, int* remHour, int* remMin, int* remSec)
{
	CAVERNnet_extendedTcpClient_c *client = new CAVERNnet_extendedTcpClient_c;

	if (client == NULL || server == NULL) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return CAVERNnet_ServerClient32_c::MEM_ALLOC_ERR;
	}

	int status = client->connectToServer(ipAddress, neighborPortNumber);
	if (status < 0) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return CAVERNnet_ServerClient32_c::CANNOT_CONNECT;
	}

	status = client->writeChar('t', CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	status = client->writeInt((serverIPNumber) ? serverIPNumber : server->getSelfIP(), CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : server->getSelfPort(), CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED; 

	// Now wait for response.
	char transferMode = ' ';
	status = client->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			*remDay = -1; *remMon = -1; *remYear = -1;
			*remHour = -1; *remMin = -1; *remSec = -1;
			return CAVERNnet_ServerClient32_c::FILE_ERROR;

		case 't': // retrieving date/time
		{
			int remdate[6];
			CAVERNnet_datapack_c packer;

			nbytes = packer.sizeof_int() * 6;
			char* read_buffer = new char[nbytes];
			status = client->read((char*) read_buffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	        packer.initUnpack(read_buffer,nbytes);  
			for (int i = 0; i < 6; i++) 
				packer.unpackInt(&(remdate[i]));
	        delete read_buffer;                     

			if (status != CAVERNnet_tcpClient_c::OK) 
				return CAVERNnet_ServerClient32_c::FILE_ERROR;

			*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
			*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
			if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
				(remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
				return CAVERNnet_ServerClient32_c::FILE_ERROR;
			}
			return CAVERNnet_ServerClient32_c::OK;
		}

		default:
			break;
	}
	return CAVERNnet_ServerClient32_c::FAILED;
}

int CAVERNnet_ServerClient32_c::initDownload(char *ipAddress, 
											char* localFileName, char* remoteFileName)
{
	CAVERNnet_datapack_c packer;
	
	client = new CAVERNnet_extendedTcpClient_c;
	if (client == NULL || server == NULL) return CAVERNnet_ServerClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, neighborPortNumber);
	if (status < 0) 
		return CAVERNnet_ServerClient32_c::CANNOT_CONNECT;

	char transferMode = 'o';
	status = client->writeChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;

	status = client->writeInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	status = client->writeInt((serverIPNumber) ? serverIPNumber : server->getSelfIP(), CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : server->getSelfPort(), CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	// Now wait for response.
	transferMode =' ';
	status = client->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) 
		return CAVERNnet_ServerClient32_c::FAILED;

	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			return CAVERNnet_ServerClient32_c::FAILED;

		case 'o': // Download 
			return CAVERNnet_ServerClient32_c::DOWNLOAD;

		default:
			return CAVERNnet_ServerClient32_c::INVALID_COMMAND;
	}
}

int CAVERNnet_ServerClient32_c::readBlock(char* dataBlock, int nBytes)
{
	if (client) {
		// Read from file and write to socket in chunks
		int status = client->read(dataBlock, &nBytes, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) 
			return CAVERNnet_ServerClient32_c::FAILED;
	
		return CAVERNnet_ServerClient32_c::OK;
	}
	return CAVERNnet_ServerClient32_c::FAILED;
}

/***************************************************************/

const int CAVERNnet_remoteFileIOServer32_c::DEFAULT_PORT = 8000;
const int CAVERNnet_remoteFileIOServer32_c::THREADED = 1;
const int CAVERNnet_remoteFileIOServer32_c::NON_THREADED = 0;
const int CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE = 30000;

int CAVERNnet_remoteFileIOServer32_c::init(int inPort, char* neighborServer, 
										   int neighborPort, int storeLocally)
{
	if (neighborServer) strcpy(neighborServerName, neighborServer);
	else neighborServerName[0] = '\0';
	store = storeLocally;
	neighborPortNumber = neighborPort;

	return server.init(inPort);
}

void CAVERNnet_remoteFileIOServer32_c::process(int threaded)
{
	CAVERNnet_extendedTcpClient_c *newClient = server.checkForNewConnections();
	if (newClient) {
        ServerData* data = new ServerData;

		data->client = newClient;
		data->neighborServerName = &neighborServerName[0];
		data->store = store;
		data->port = neighborPortNumber;
		data->server = &server;

		if (threaded== CAVERNnet_remoteFileIOServer32_c::THREADED) {
			CAVERNts_thread_c newThread; // = new CAVERNnet_thread_c();
			newThread.create(threadHandler, (void*) data);
		} 
		else threadHandler((void*) data);
	}
}

#define CLOSE_REMOVE_AND_RETURN 	if (readFile) fclose(readFile); \
									if (writeFile) fclose(writeFile); \
									if (downloadClient) delete downloadClient; \
									delete data; \
									return NULL; 

void* CAVERNnet_remoteFileIOServer32_c::threadHandler(void* inData)
{
	ServerData* data = (ServerData*) inData;
	if (data == NULL) return NULL;

	CAVERNnet_datapack_c packer;
	CAVERNnet_extendedTcpClient_c *newClient = ((ServerData *) data)->client;
	char* neighborServerName = ((ServerData*) data)->neighborServerName;
	int store = ((ServerData*) data)->store;
	int neighborPort = ((ServerData*) data)->port;
	CAVERNnet_tcpServer_c *server = ((ServerData *) data)->server;
	int nbytes;

	// Check the command
	char transferMode;
	int status = newClient->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {
	    delete data; return NULL;
	}
    int filenamelen;
    char* filename;
    
	if (transferMode != 'f')    //reading pathname len and pathname is taken care of at a later stage for 'f'
    {
        // Check filename size
	    status = newClient->readInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	    if (status != CAVERNnet_tcpClient_c::OK) { delete data; return NULL; }

	    // Check filename
	    filename = new char[filenamelen];
	    if (filename == NULL) { delete data; return NULL;}
	    status = newClient->read(filename, &filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	    if (status != CAVERNnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }
	    filename[filenamelen - 1]='\0';
    }
	char dataBlock[CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE];

	// If handling check for return date/time stamp for the request from the client
	if (transferMode == 'c') {
		int thedate[6];

		// If file does not exist time stamp will return -1s.
		status = CAVERNnet_remoteFileIOClient32_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);

		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != CAVERNnet_remoteFileIOClient32_c::OK && neighborServerName[0]) {
#ifdef DBG_MSG
			printf("server->getSelfPort = %d, neighborPort = %d\n", server->getSelfPort(), neighborPort);
#endif
			if (server->getSelfIP() == getHostIPNumber(neighborServerName) 
							&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
				printf("trying to connecting to itself. server processing 'c'\n");
#endif
				status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
							    
		    CAVERNnet_ServerClient32_c client(server);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != CAVERNnet_ServerClient32_c::OK) {
				// Send back command 'n'
				status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 'c'
		status = newClient->writeChar('c', CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time of file - also add the local time stamp to the stream -for latency calculations
		nbytes = 6 * packer.sizeof_int() + packer.sizeof_double();
		char* send_buffer = new char[nbytes];
        double timeStamp = newClient->getTimeInSecs();
        
        packer.initPack(send_buffer,nbytes);
		for (int i = 0; i < 6; i++) packer.packInt(thedate[i]);
		packer.packDouble(timeStamp);
        status = newClient->write((char*) send_buffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling check for return date/time stamp for the request from a server
	if (transferMode == 't') {
		// Get server IP
		int serverIP;
		status = newClient->readInt(serverIP, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

		// get port number 
		int serverPort;
		newClient->readInt(serverPort, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

/*		// Check whether we are getting back the request this server sent before.
		if (serverIP == server->getSelfIP() && serverPort == server->getSelfPort()) { 
#ifdef DBG_MSG
			printf("cycle detected. server processing 't'\n");
#endif
			newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
			delete filename; delete data; return NULL;
		}*/
		
		int thedate[6];

		// If file does not exist time stamp will return -1s.
		status = CAVERNnet_remoteFileIOClient32_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);

		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != CAVERNnet_remoteFileIOClient32_c::OK && neighborServerName[0]) {
#ifdef DBG_MSG
			printf("serverPort = %d, neighborPort = %d\n", serverPort, neighborPort);
#endif
			if (serverIP == getHostIPNumber(neighborServerName) 
							&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
				printf("Cycle detected. Server processing 't'\n");
#endif
				status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}

		    CAVERNnet_ServerClient32_c client(server, serverIP, serverPort);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != CAVERNnet_ServerClient32_c::OK) {
				newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 't'
		status = newClient->writeChar('t', CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time stamp of file
		nbytes = 6*packer.sizeof_int();
		char* send_buffer = new char[nbytes];
        packer.initPack(send_buffer,nbytes);
		for (int i = 0; i < 6; i++) packer.packInt(thedate[i]);
		status = newClient->write((char*) send_buffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling upload
	if (transferMode == 'u') {
		// Get datasize
		int dataSize;
		status = newClient->readInt(dataSize, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

		FILE *writeFile = fopen(filename,"wb");
		if (writeFile == NULL) {
			delete filename; delete data; return NULL;
		}

		// Read from file and write to socket in 8K chunks
		int numChunks = dataSize / CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
		for (int i = 0; i < numChunks; i++) {
			nbytes = CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
			status = newClient->read(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				fclose(writeFile); delete filename; delete data; return NULL;
			}

			fwrite(dataBlock, nbytes, 1, writeFile);
		}
		int remain = dataSize - (numChunks * CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			status = newClient->read(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				fclose(writeFile); delete filename; delete data; return NULL;
			}

			nbytes = remain;
			fwrite(dataBlock, nbytes, 1, writeFile);
		}
		fclose(writeFile);
		delete filename;
		delete data;
		return NULL;
	}
#if (defined(__sgi)|| defined(linux) || defined(sun))
    if (transferMode == 'f')
    {
        //protocol from client for a file list-> 'f' + pathnamelen + pathname + recursion flag
        int pathnamelen;
        status = newClient->readInt(pathnamelen, CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) {delete data; return NULL; }
        char* pathname = new char[pathnamelen]; 
        
        nbytes = pathnamelen;
        status = newClient->read((char*)pathname, &nbytes, CAVERNnet_tcpClient_c::BLOCKING); 
        if (status != CAVERNnet_tcpClient_c::OK) { delete pathname; delete data; return NULL;}
        
        int doRecursion;
        status = newClient->readInt(doRecursion, CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) { delete data; return NULL; }
       
        int linkTraversal;
        status = newClient->readInt(linkTraversal, CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) { delete data; return NULL; }
        
        //send reply back to the client - 'f' for command identity
        status = newClient->writeChar('f', CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) {delete data; return NULL;}
        
        //build the file list under the given path
        CAVERNmisc_fileList_c list;
        int result = list.buildFileList(pathname, doRecursion, linkTraversal);
        delete pathname; pathname = NULL;
        int noObjects;
        if (result == CAVERNmisc_fileList_c::FAILED)    //A file error occured - write error message to client
        {
            noObjects = -1;   //to indicate failure
            status = newClient->writeInt(noObjects, CAVERNnet_tcpClient_c::BLOCKING);
            delete data; 
            return NULL;    
        }
        noObjects = list.getNoFiles();
        
        status = newClient->writeInt(noObjects, CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) {delete data; return NULL;}
        
        //get the datapacking size
        int packingSize = list.getDataPackingSize();
        double timeStamp = newClient->getTimeInSecs();
        
        //pack the datasize and the time stamp and send it to the client
        nbytes = packer.sizeof_int() +packer.sizeof_double();
        char* buffer = new char[nbytes];
        packer.initPack(buffer, nbytes);
        packer.packInt(packingSize);
        packer.packDouble(timeStamp);
        status = newClient->write((char*)buffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
        if (status != CAVERNnet_tcpClient_c::OK) {delete buffer; delete data; return NULL;}
        delete buffer;
        
        //pack the file objects and send them across
        char* listBuffer = new char[packingSize];
        CAVERNnet_datapack_c* listPacker = new CAVERNnet_datapack_c;
        listPacker->initPack(listBuffer, packingSize);
        
        list.packFileList(listPacker);
        nbytes = packingSize;
        status = newClient->write((char*)listBuffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);        
        delete listBuffer; listBuffer = NULL;
        delete listPacker;
        delete data;
        return NULL;
    }
#endif
    
	// If handling download... 
	if (transferMode == 'd') {
		int fileSize;

		FILE *readFile = NULL, *writeFile = NULL;
		CAVERNnet_ServerClient32_c* downloadClient = NULL;
	
		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
#ifdef DBG_MSG
				printf("server->getSelfPort = %d, neighborPort = %d\n", server->getSelfPort(), neighborPort);
#endif
				if (server->getSelfIP() == getHostIPNumber(neighborServerName) 
								&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
					printf("Trying to connecting to itself. server processing 'd'\n");
#endif
					status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}

				downloadClient = new CAVERNnet_ServerClient32_c(server);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename) !=
					                                  CAVERNnet_ServerClient32_c::DOWNLOAD) {
					newClient->writeChar('i', CAVERNnet_tcpClient_c::BLOCKING);
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		status = newClient->writeChar('d', CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek(readFile, 0, SEEK_END);
			fileSize = ftell(readFile);
			fseek(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			status = downloadClient->getFileSize(fileSize);
			if (status != CAVERNnet_ServerClient32_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file along with the timestamp, for latency calculations
        nbytes = packer.sizeof_int() + packer.sizeof_double();
        char* sendBuffer = new char[nbytes];
        packer.initPack(sendBuffer, nbytes);
        double timeStamp = newClient->getTimeInSecs();
        packer.packInt(fileSize);            
        packer.packDouble(timeStamp);
        	status = newClient->write((char*) sendBuffer, &nbytes, CAVERNnet_tcpClient_c :: BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int numChunks = fileSize / (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
		for(int i = 0; i < numChunks; i++) {
			nbytes = CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, nbytes, 1, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != CAVERNnet_ServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = fileSize - (numChunks * (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, nbytes,1,readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != CAVERNnet_ServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	// If handling download for the request from a server... 
	if (transferMode == 'o') {
		int fileSize;
		FILE *readFile = NULL, *writeFile = NULL;
		CAVERNnet_ServerClient32_c* downloadClient = NULL;

		// Get server name's length
		int serverIP;
		if (newClient->readInt(serverIP, CAVERNnet_tcpClient_c::BLOCKING) != CAVERNnet_tcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}

		// get port number 
		int serverPort;
		if (newClient->readInt(serverPort, CAVERNnet_tcpClient_c::BLOCKING) != CAVERNnet_tcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}
		
		// Check whether we are getting back the request this server sent before.
/*		if (serverIP == server->getSelfIP() && serverPort == server->getSelfPort()) { 
			printf("cycle detected\n");
			newClient->writeChar('i', CAVERNnet_tcpClient_c::BLOCKING);
			delete data;
			return NULL;
		}*/

		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
#ifdef DBG_MSG
				printf("serverPort = %d, neighborPort = %d\n", serverPort, neighborPort);
#endif
				if (serverIP == getHostIPNumber(neighborServerName) 
								&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
					printf("Cycle detected in server processing 'o'\n");
#endif
					status = newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}

				downloadClient = new CAVERNnet_ServerClient32_c(server, serverIP, serverPort);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename) !=
					CAVERNnet_ServerClient32_c::DOWNLOAD) {
					newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		if (newClient->writeChar('o', CAVERNnet_tcpClient_c::BLOCKING) != CAVERNnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek(readFile, 0, SEEK_END);
			fileSize = ftell(readFile);
			fseek(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			if (downloadClient->getFileSize(fileSize) != CAVERNnet_ServerClient32_c::OK) {
				newClient->writeChar('n', CAVERNnet_tcpClient_c::BLOCKING);
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file
		if (newClient->writeInt(fileSize, CAVERNnet_tcpClient_c::BLOCKING) != CAVERNnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int numChunks = fileSize / (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
		for(int i = 0; i < numChunks; i++) {
			nbytes = CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, nbytes, 1, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != CAVERNnet_ServerClient32_c::OK)	{
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = fileSize - (numChunks * (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, nbytes,1,readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != CAVERNnet_ServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	return NULL;
}
#undef CLOSE_REMOVE_AND_RETURN 	

int CAVERNnet_remoteFileIOServer32_c::getHostIPNumber(char* hostNameOrIP)
{
	struct hostent* hostEnt;

	unsigned ip = inet_addr(hostNameOrIP);
	if ((ip != INADDR_NONE) || (strcmp(hostNameOrIP, "255.255.255.255") == 0))
		return (int) ip;
	else {
		hostEnt = gethostbyname(hostNameOrIP);
		if (hostEnt) 
			return *((int*) hostEnt->h_addr);
		else {
		    printf("Invalid address is given: %s\n", hostNameOrIP);
			return 0;
		}
	}
}

/***************************************************************/

const int CAVERNnet_remoteFileIOClient32_c::OK = 1;
const int CAVERNnet_remoteFileIOClient32_c::FAILED = 0;
const int CAVERNnet_remoteFileIOClient32_c::CANNOT_CONNECT = 2;
const int CAVERNnet_remoteFileIOClient32_c::CANNOT_READ_LOCAL_FILE = 3;
const int CAVERNnet_remoteFileIOClient32_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int CAVERNnet_remoteFileIOClient32_c::INVALID_COMMAND = 4;
const int CAVERNnet_remoteFileIOClient32_c::MEM_ALLOC_ERR = 5;
const int CAVERNnet_remoteFileIOClient32_c::FILE_ERROR = 6;
const int CAVERNnet_remoteFileIOClient32_c::EARLIER = 7;
const int CAVERNnet_remoteFileIOClient32_c::LATER = 8;
const int CAVERNnet_remoteFileIOClient32_c::SAME = 9;

CAVERNnet_remoteFileIOClient32_c::CAVERNnet_remoteFileIOClient32_c()
{
	remotePort = CAVERNnet_remoteFileIOServer32_c::DEFAULT_PORT;
}

void CAVERNnet_remoteFileIOClient32_c::setRemotePort(int port) {remotePort = port;}

int CAVERNnet_remoteFileIOClient32_c::getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, int *remMon, int *remYear, int* remHour, int* remMin, int* remSec)
{
	CAVERNnet_extendedTcpClient_c *client = new CAVERNnet_extendedTcpClient_c;
	CAVERNnet_datapack_c packer;
	char *read_buffer;
	int nbytes;

	if (client == NULL) return CAVERNnet_remoteFileIOClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status < 0) {
		return CAVERNnet_remoteFileIOClient32_c::CANNOT_CONNECT;
	}

	status = client->writeChar('c', CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data sent
    pmonitor.incrementDataSent(sizeof(char));
    
	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;  
    nbytes = packer.sizeof_int();
	status = client->writeInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
	
    // Now wait for response.
	char transferMode =' ';
	status = client->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'c') {
		int remdate[6];
        double remoteTimeStamp;
        
		nbytes = packer.sizeof_int() * 6 + packer.sizeof_double();
		read_buffer = new char[nbytes];
		status = client->read((char*) read_buffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
        packer.initUnpack(read_buffer,nbytes);  
		for (int i = 0; i < 6; i++) 
			packer.unpackInt(&(remdate[i]));
        packer.unpackDouble(&remoteTimeStamp);
        delete read_buffer;                     

		if (status != CAVERNnet_tcpClient_c::OK) {
			client->close(); delete client;	return CAVERNnet_remoteFileIOClient32_c::FILE_ERROR;
		}

       //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //Calculations for performance monitoring
        double localTime = client->getTimeInSecs();
        double latency = fabs(localTime - remoteTimeStamp);
        pmonitor.setInstantLatency(latency);

       //set IP and port info for the performance monitoring statistics        
        char selfIp[256], remoteIp[256];
        
        client->getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
   
		*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
		*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
		if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
		    (remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
			client->close(); delete client;	return CAVERNnet_remoteFileIOClient32_c::FILE_ERROR;
		}
		client->close(); delete client;
		return CAVERNnet_remoteFileIOClient32_c::OK;
	}
	else {
		int result = CAVERNnet_remoteFileIOClient32_c::FAILED;

		if (transferMode == 'i')
			result = CAVERNnet_remoteFileIOClient32_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = CAVERNnet_remoteFileIOClient32_c::FILE_ERROR;

		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return result;
	}
}

int CAVERNnet_remoteFileIOClient32_c::upload(char *ipAddress, char* localFileName, char* remoteFileName)
{
	CAVERNnet_extendedTcpClient_c *client = new CAVERNnet_extendedTcpClient_c;
	CAVERNnet_datapack_c packer;

	if (client == NULL) return CAVERNnet_remoteFileIOClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status < 0) {
		return CAVERNnet_remoteFileIOClient32_c::CANNOT_CONNECT;
	}

	char transferMode = 'u';

	status = client->writeChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data sent
    int nbytes = sizeof(char);
    pmonitor.incrementDataSent(nbytes);

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //increment data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
	
    // Send data size
	// Find length of file.
	FILE *readFile = fopen(localFileName, "rb");

	if (readFile==NULL) {
		client->close(); delete client;
		return CAVERNnet_remoteFileIOClient32_c::CANNOT_READ_LOCAL_FILE;
	}

	fseek(readFile, 0, SEEK_END);
	int fileSize = ftell(readFile);
	fseek( readFile , 0L, SEEK_SET);

	status = client->writeInt(fileSize, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);
	// Send data
	char dataBlock[CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE];

	int numChunks = fileSize / CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;

	for (int i = 0; i < numChunks; i++) {
		fread(dataBlock, CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE, 1, readFile);
		nbytes = CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
		status = client->write(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return CAVERNnet_remoteFileIOClient32_c::FAILED;}

        //Increment the data sent
        pmonitor.incrementDataSent(nbytes);
	}
    
	int remain = fileSize - (numChunks * CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE);
	if (remain) {
		fread(dataBlock, remain,1,readFile);
		status = client->write(dataBlock, &remain, CAVERNnet_tcpClient_c::BLOCKING);
		if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return CAVERNnet_remoteFileIOClient32_c::FAILED;}

        //Increment the data sent
        pmonitor.incrementDataSent(remain);
	}
    //set IP and port info for the performance monitoring statistics        
    char selfIp[256], remoteIp[256];
    
    client->getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
   
	fclose(readFile);
	client->close();
	delete client;
	return CAVERNnet_remoteFileIOClient32_c::OK;
}

#if (defined(__sgi)|| defined(linux) || defined(sun))

int CAVERNnet_remoteFileIOClient32_c::getRemoteFileList(char* ipAddress, char* remotePath, CAVERNmisc_fileInfo_c*& fileList, int* noObjects, int doRecursion, int linkTraversal)
{
    CAVERNnet_extendedTcpClient_c *client = new CAVERNnet_extendedTcpClient_c;
    CAVERNnet_datapack_c packer;
    if (client == NULL) return CAVERNnet_remoteFileIOClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
    if (status < 0) {
        client->close(); delete client;
        return CAVERNnet_remoteFileIOClient32_c::CANNOT_CONNECT;
        }
    int nbytes;
    char transferMode = 'f';                
    status = client->writeChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) 
    {
        client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;
    }
    //Increment the data sent
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataSent(nbytes);

	// Send remotePathname length
	int pathnamelen = strlen(remotePath) + 1;
	status = client->writeInt(pathnamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data sent
	nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send pathname
	nbytes = pathnamelen;
	status = client->write(remotePath, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

     //Increment the data sent
    pmonitor.incrementDataSent(nbytes);	
	
    // Send the recursion flag
    nbytes = packer.sizeof_int();
    status = client->writeInt(doRecursion, CAVERNnet_tcpClient_c::BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
    
    //Send the linkTraversal flag
    nbytes = packer.sizeof_int();
    status = client->writeInt(linkTraversal, CAVERNnet_tcpClient_c::BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
     
    transferMode =' ';
	status = client->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'f') {    
        //unpack no_objects(int), datasize(int) + time stamp(double) and unpack the file list
    int noFiles;
    status = client->readInt(noFiles, CAVERNnet_tcpClient_c::BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data read
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataRead(nbytes);
    if (noFiles == -1)    
        {
            printf("\nFile error on the remote server - please check the path nae");
            return FILE_ERROR;
        }
    fileList = new CAVERNmisc_fileInfo_c[noFiles];
    *noObjects = noFiles;
    int dataSize;
    double remoteTimeStamp;
    nbytes = packer.sizeof_int() + packer.sizeof_double();
    char* readBuffer = new char[nbytes];
    
    status = client->read((char*) readBuffer, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; delete readBuffer; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //unpack the datasize (along with the time stamp to find out Inst. latency - for performance monitoring)
    packer.initUnpack(readBuffer,nbytes);
    packer.unpackInt(&dataSize);   
    packer.unpackDouble(&remoteTimeStamp);
    delete readBuffer; readBuffer = NULL; 
       
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

    //Calculations for performance monitoring
    double localTime = client->getTimeInSecs();
    double latency = fabs(localTime - remoteTimeStamp);
    pmonitor.setInstantLatency(latency); 
    nbytes = dataSize;
    readBuffer = new char[nbytes];
    if (readBuffer == NULL)
        {
            printf("\nMemory allocation error\n");
            return MEM_ALLOC_ERR;
        }       
    status = client->read((char*) readBuffer, &nbytes, CAVERNnet_tcpClient_c :: BLOCKING);
    if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; delete readBuffer; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);
    //set IP and port info for the performance monitoring statistics
    char selfIp[256], remoteIp[256];        
    client->getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort()); 
    
    CAVERNmisc_fileList_c list;
    CAVERNnet_datapack_c* listPacker = new  CAVERNnet_datapack_c;
    listPacker->initUnpack(readBuffer, dataSize);
    if (list.unpackfileList(listPacker, remotePath, fileList, noFiles) == CAVERNmisc_fileList_c::FAILED)
        return FAILED;

    //list.displayFileList();
    delete readBuffer;   readBuffer =NULL;
    delete listPacker;

    } //endif (transferMode == 'f')
    
    client->close(); delete client;
    return OK;
}

#endif

int CAVERNnet_remoteFileIOClient32_c::download(char *ipAddress, char* localFileName, char* remoteFileName)
{
	CAVERNnet_extendedTcpClient_c *client = new CAVERNnet_extendedTcpClient_c;
	CAVERNnet_datapack_c packer;

	if (client == NULL) return CAVERNnet_remoteFileIOClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status < 0) {
		return CAVERNnet_remoteFileIOClient32_c::CANNOT_CONNECT;
	}

	int nbytes;
	char transferMode = 'd';
	status = client->writeChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;
    }
    //Increment the data sent
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataSent(nbytes);

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}
    
    //Increment the data sent
	nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

     //Increment the data sent
    pmonitor.incrementDataSent(nbytes);	
	
    transferMode =' ';
	status = client->readChar(transferMode, CAVERNnet_tcpClient_c::BLOCKING);
	if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'd') {
		int dataSize;
        double remoteTimeStamp;
		nbytes = packer.sizeof_int() + packer.sizeof_double();
        char* read_buffer = new char[nbytes];
        status = client->read((char*) read_buffer, &nbytes, CAVERNnet_tcpClient_c :: BLOCKING);
		
		if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; return CAVERNnet_remoteFileIOClient32_c::FAILED;}

        packer.initUnpack(read_buffer,nbytes);
        packer.unpackInt(&dataSize);   
        packer.unpackDouble(&remoteTimeStamp);
        delete read_buffer; 
        
        //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //Calculations for performance monitoring
        double localTime = client->getTimeInSecs();
        double latency = fabs(localTime - remoteTimeStamp);
        pmonitor.setInstantLatency(latency);
 
		FILE *writeFile = fopen(localFileName,"wb");
		if (writeFile == NULL) {
			client->close(); delete client;
			return CAVERNnet_remoteFileIOClient32_c::CANNOT_WRITE_LOCAL_FILE;
		}

		// Read from file and write to socket in chunks
		int numChunks = dataSize / (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
		char dataBlock[CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE];

		for (int i = 0; i < numChunks; i++) {
			nbytes = CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE;
			status = client->read(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; fclose(writeFile); return CAVERNnet_remoteFileIOClient32_c::FAILED;}

            //Increment the data read
            pmonitor.incrementDataRead(nbytes);

			fwrite(dataBlock, nbytes, 1, writeFile);
		}

		int remain = dataSize - (numChunks * (int) CAVERNnet_remoteFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			status = client->read(dataBlock, &nbytes, CAVERNnet_tcpClient_c::BLOCKING);
			if (status != CAVERNnet_tcpClient_c::OK) {client->close(); delete client; fclose(writeFile); return CAVERNnet_remoteFileIOClient32_c::FAILED;}

            //Increment the data read
            pmonitor.incrementDataRead(nbytes);

			nbytes = remain;
			fwrite(dataBlock, nbytes, 1, writeFile);
		}
        
        //set IP and port info for the performance monitoring statistics
        char selfIp[256], remoteIp[256];        
        client->getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());

		fclose(writeFile);
		client->close();
		delete client;
		return CAVERNnet_remoteFileIOClient32_c::OK;
	} 
	else {
		int result;

		if (transferMode == 'i') 
			result = CAVERNnet_remoteFileIOClient32_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = CAVERNnet_remoteFileIOClient32_c::FILE_ERROR;

		client->close();
		delete client;
		return result;
	}
}

// Given a filename return its modification time
int CAVERNnet_remoteFileIOClient32_c::getLocalDateTime(char* filename,
		 int *day,
		 int *month,
		 int *year,
		 int *hour,
		 int *minute,
		 int *second)
{

	struct stat buf;

	*day = *month = *year = *hour = *minute = *second = -1;

	if (stat(filename, &buf) != 0) return CAVERNnet_remoteFileIOClient32_c::FILE_ERROR;

	time_t tim;

#ifdef __sgi                      
    tim = buf.st_mtim.tv_sec; 
#else                      
    tim = buf.st_mtime;       
#endif                            
                                  
	struct tm *gmt;
	gmt = gmtime(&tim);

	*day = gmt->tm_mday;
	*month = gmt->tm_mon + 1;
	*year = gmt->tm_year + 1900;
	*hour = gmt->tm_hour;
	*minute = gmt->tm_min;
	*second = gmt->tm_sec;

	return CAVERNnet_remoteFileIOClient32_c::OK;
}

int CAVERNnet_remoteFileIOClient32_c::compareDateTime(int day1, int month1, int year1,
						      int hour1, int minute1, int second1,
						      int day2, int month2, int year2,
						      int hour2, int minute2, int second2)
{
	if (year1 < year2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (year1 > year2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	if (month1 < month2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (month1 > month2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	if (day1 < day2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (day1 > day2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	if (hour1 < hour2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (hour1 > hour2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	if (minute1 < minute2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (minute1 > minute2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	if (second1 < second2) return CAVERNnet_remoteFileIOClient32_c::EARLIER;
	if (second1 > second2) return CAVERNnet_remoteFileIOClient32_c::LATER;

	return CAVERNnet_remoteFileIOClient32_c::SAME;
}


void CAVERNnet_remoteFileIOClient32_c::showStats(char* streamInfo, char* comment)        
{
    pmonitor.showStats(streamInfo, comment);    
}

int CAVERNnet_remoteFileIOClient32_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}


int CAVERNnet_remoteFileIOClient32_c::sendStats(char* streamInfo, char* comment)
{
    return pmonitor.sendStats(streamInfo, comment);    
}

int CAVERNnet_remoteFileIOClient32_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void CAVERNnet_remoteFileIOClient32_c::exitSendStats()
{
    pmonitor.exitSendStats();
}



