/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERNnet_tcp_c.hxx"

#ifndef WIN32
#include <netinet/tcp.h>
#endif

const int CAVERNnet_tcpServer_c::OK = 1;
const int CAVERNnet_tcpServer_c::FAILED = 0;
#ifdef sun
long gethostid(void);
#endif

#ifdef WIN32
#define MAXHOSTNAME   128   /* Maximum length of a DNS hostname */
#define DEFAULT_BUFFER_SIZE 32766 /* Default socket bufffer size */

// This function is excerpted from "Windows Sockets Network Programming"
// by Bob Quinn and Dave Shute published from Addison-Wesley.
static long gethostid(void)
{
	char szLclHost [MAXHOSTNAME];
	LPHOSTENT lpstHostent;
	SOCKADDR_IN stLclAddr;
	SOCKADDR_IN stRmtAddr;
	int nAddrSize = sizeof(SOCKADDR);
	SOCKET hSock;
	int nRet;

	/* Init local address (to zero) */
	stLclAddr.sin_addr.s_addr = INADDR_ANY;

	/* Get the local hostname */
	nRet = gethostname(szLclHost, MAXHOSTNAME); 
	if (nRet != SOCKET_ERROR) {
		/* Resolve hostname for local address */
		lpstHostent = gethostbyname((LPSTR)szLclHost);
		if (lpstHostent)
			stLclAddr.sin_addr.s_addr = *((u_long FAR*) (lpstHostent->h_addr));
	} 

	/* If still not resolved, then try second strategy */
	if (stLclAddr.sin_addr.s_addr == INADDR_ANY) {
		/* Get a UDP socket */
		hSock = socket(AF_INET, SOCK_DGRAM, 0);
		if (hSock != INVALID_SOCKET)  {
			/* Connect to arbitrary port and address (NOT loopback) */
			stRmtAddr.sin_family = AF_INET;
			stRmtAddr.sin_port   = htons(IPPORT_ECHO);
			stRmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
			nRet = connect(hSock, (LPSOCKADDR)&stRmtAddr,
						   sizeof(SOCKADDR));
			if (nRet != SOCKET_ERROR) {
				/* Get local address */
				getsockname(hSock, (LPSOCKADDR)&stLclAddr, 
							(int FAR*)&nAddrSize);
			}
			closesocket(hSock);   /* we're done with the socket */
		}
	}
	return (stLclAddr.sin_addr.s_addr);
}
#endif

CAVERNnet_tcpServer_c::CAVERNnet_tcpServer_c()
{
	timeOutPeriod = CAVERNnet_tcpClient_c::DEFAULT_TIME_OUT;
}

void CAVERNnet_tcpServer_c::close() {
#ifdef WIN32
	if (sockfd > 0) ::closesocket(sockfd);
#else
	if (sockfd > 0) ::close(sockfd);
#endif
	sockfd = 0;
}

int CAVERNnet_tcpServer_c::init(int port)
{
	int optlen;

	/***** Open a TCP socket (an Internet stream socket). */
	/* Notice we want Internet, stream socket and not datagram */
#ifdef WIN32
	if ((sockfd = socket(AF_INET,SOCK_STREAM,0)) == INVALID_SOCKET) {
		return CAVERNnet_tcpServer_c::FAILED;
	}
#else
	if ((sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
		return CAVERNnet_tcpServer_c::FAILED;
	}
#endif
	int sendbuff;

	/* Make sure the buffer is big enough */

	sendbuff = DEFAULT_BUFFER_SIZE;

	optlen = sizeof(sendbuff);
	if (setsockopt(sockfd,SOL_SOCKET, SO_SNDBUF, (char *) &sendbuff, optlen) < 0)
		printf("CAVERNnet_tcpServer_c::init: Cannot create a large enough socket buffer.\n");


	/* Set no delay to stop TCP/IP from buffering */
	int noDelay = 0;// Set to 0 for no buffering. Set to 1 for buffering.
	optlen = sizeof(noDelay);

	if (setsockopt(sockfd,IPPROTO_TCP,TCP_NODELAY, (char *) &noDelay, optlen) < 0) 
		printf("CAVERNnet_tcpServer_c::init: Cannot set no delay, you may experience sluggish performance.\n");

	/* Use this to try to eliminate the problem of being unable to rebind to a static port */
	int reuseFlag = 1;
	optlen = sizeof(reuseFlag);
	if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR, (char *) &reuseFlag, optlen) < 0) 
		printf("CAVERNnet_tcpServer_c::init: Cannot reuse port address.\n");
	/***** Bind our local address so that the client can send to us. */

	struct linger lingerData;
	lingerData.l_onoff = 1;
	lingerData.l_linger = 1000;

	optlen = sizeof(lingerData);
	if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER, (char *) &lingerData, optlen) < 0)
		printf("CAVERNnet_tcpServer_c::init: setsockopt: SO_LINGER failed.\n");

	/* Clear out the structure */
	memset((char *) &serverInfo, 0, sizeof(serverInfo));

	serverInfo.sin_family = AF_INET;
	serverInfo.sin_addr.s_addr = htonl(INADDR_ANY);
	serverInfo.sin_port = htons(port);

	if (bind(sockfd, (struct sockaddr *) &serverInfo, sizeof(serverInfo)) < 0) {
		printf("CAVERNnet_tcpServer_c::init: Can't bind local address.\n");
		return CAVERNnet_tcpServer_c::FAILED;
	}

	/* Listen for a connection. Allow queuing for 5 connection requests.
	 */
	listen(sockfd,5);

	/* Non blocking from now on. This is necessary in order for checkForNewConnections() to work.But this seems to be the main thing that is screwing up
	   with threads.
	 */
	//	fcntl(sockfd,F_SETFL,FNDELAY);

	return CAVERNnet_tcpServer_c::OK;
}

unsigned int CAVERNnet_tcpServer_c::getSelfIP()
{
#ifdef linux
    if (serverInfo.sin_addr.s_addr) return serverInfo.sin_addr.s_addr;
    else {
        unsigned short a1, a2;
        unsigned long ip, temp;

        ip = gethostid();
        a1 = (unsigned short) (ip >> 16);
        a2 = (unsigned short) ip;
        temp = (unsigned long) a2;
        temp <<= 16;
        temp |= a1;
        return temp;            
    } 
#else
    return (serverInfo.sin_addr.s_addr) ? serverInfo.sin_addr.s_addr : gethostid();
#endif
}

void CAVERNnet_tcpServer_c :: getSelfIP(char *name)
{
        ipNumToString(getSelfIP(), name);
}

const int CAVERNnet_tcpClient_c::BLOCKING = 1;
const int CAVERNnet_tcpClient_c::NON_BLOCKING= 2;
const int CAVERNnet_tcpClient_c::NO_TIME_OUT = -1;
const int CAVERNnet_tcpClient_c::DEFAULT_TIME_OUT = NO_TIME_OUT; 
const int CAVERNnet_tcpClient_c::SOCKET_NOT_OPEN = -1;
const int CAVERNnet_tcpClient_c::CONNECTION_TERMINATED = -2;
const int CAVERNnet_tcpClient_c::NON_BLOCKING_HAS_NO_DATA =-3;
const int CAVERNnet_tcpClient_c::TIMED_OUT = -4;
const int CAVERNnet_tcpClient_c::OK = 1;
const int CAVERNnet_tcpClient_c::NON_BLOCKING_NOT_READY_TO_WRITE = -6;
const int CAVERNnet_tcpClient_c::NOT_READY = -7;
const int CAVERNnet_tcpClient_c::READY_TO_READ = -8;
const int CAVERNnet_tcpClient_c::READY_TO_WRITE = -9;
const int CAVERNnet_tcpClient_c::READY_TO_READ_AND_WRITE = -10;
const int CAVERNnet_tcpClient_c::READ_BUFFER_SIZE = -11;  
const int CAVERNnet_tcpClient_c::WRITE_BUFFER_SIZE = -12;

CAVERNnet_tcpClient_c *CAVERNnet_tcpServer_c::checkForNewConnections()
{
	struct sockaddr_in clientInfo;
	int clilen;

	// Perform a non-blocking check to see if socket is ready for
	// an incoming connection.
	fd_set read_set;
	struct timeval timeout;
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	int ready_fd;

	FD_ZERO(&read_set);
	FD_SET(sockfd, &read_set);
	ready_fd = select( sockfd+1, &read_set, NULL, NULL, &timeout);

	// If it is not ready then return.
	if ((ready_fd <= 0)  ||  (!FD_ISSET( sockfd, &read_set))) return NULL;

	clilen = sizeof(clientInfo);
	int newsockfd = accept(sockfd,
			       (struct sockaddr *) &clientInfo,
			       &clilen);
	if (newsockfd > 0) {
		struct linger lingerData;
		lingerData.l_onoff = 1;
		lingerData.l_linger = 1000;

		int optlen = sizeof(lingerData);
	
		if (setsockopt(newsockfd,SOL_SOCKET,SO_LINGER, (char *) &lingerData, optlen) < 0)
			printf("CAVERNnet_tcpServer_c::checkForNewConnections: setsockopt: SO_LINGER failed.\n");

		CAVERNnet_tcpClient_c *clientObj;

		clientObj = new CAVERNnet_tcpClient_c;
		clientObj->timeOutPeriod = timeOutPeriod;
		clientObj->clientInfo = clientInfo;
		clientObj->clientSockFd = newsockfd;
		//		clientObj->makeNonBlocking();
		return clientObj;
	} else return NULL;
}

/*
 * For the blockingtype parameter:
 * NEVER_TIMEOUT_BLOCKING means the timeout period is bypassed so that it will
 * continue to read forever unless there is a break in the connection.
 *
 * NON_BLOCKING means that if there is no data to be read this call will
 * return immediately.
 *
 * BLOCKING means that it will block waiting for data or until the timeout
 * period expires. 
 * Change the timeout period (in seconds) by changing the timeoutPeriod
 * member variable.
 *
 */

void CAVERNnet_tcpClient_c::setSockOptions(int option, int buffersize)
{
	/*Change the socket buffer size if default is not to be used*/	
	int optlen = sizeof(buffersize);

	if (option == CAVERNnet_tcpClient_c::READ_BUFFER_SIZE)
		if (setsockopt(clientSockFd,SOL_SOCKET, SO_RCVBUF, (char *) &buffersize, optlen) < 0)
			printf("CAVERNnet_tcpClient_c::setSockOptions: Cannot create a large enough socket buffer for Read.\n");

	if (option == CAVERNnet_tcpClient_c::WRITE_BUFFER_SIZE)
		if (setsockopt(clientSockFd,SOL_SOCKET, SO_SNDBUF, (char *) &buffersize, optlen) < 0)
			printf("CAVERNnet_tcpClient_c::setSockOptions: Cannot create a large enough socket buffer for Write.\n");

}

int CAVERNnet_tcpClient_c::isReadyToRead()
{
	struct timeval timeoutDelay;
	fd_set readfds;

	timeoutDelay.tv_sec = 0;
	timeoutDelay.tv_usec = 0;

	FD_ZERO(&readfds);
	FD_SET(clientSockFd, &readfds);
	select(clientSockFd+1,&readfds,NULL,NULL, &timeoutDelay);
	
	if (!(FD_ISSET(clientSockFd, &readfds)))
		return NOT_READY;
	else
		return READY_TO_READ;
}

int CAVERNnet_tcpClient_c::isReadyToWrite()
{
	struct timeval timeoutDelay;
	fd_set writefds;

	timeoutDelay.tv_sec = 0;
	timeoutDelay.tv_usec = 0;

	FD_ZERO(&writefds);
	FD_SET(clientSockFd, &writefds);
	select(clientSockFd+1,NULL,&writefds,NULL, &timeoutDelay);
	
	if (!(FD_ISSET(clientSockFd, &writefds)))
		return NOT_READY;
	else
		return READY_TO_WRITE;
}

int CAVERNnet_tcpClient_c::isReady()
{
	struct timeval timeoutDelay;
	fd_set writefds, readfds;

	timeoutDelay.tv_sec = 0;
	timeoutDelay.tv_usec = 0;

	FD_ZERO(&writefds);
	FD_ZERO(&readfds);
	FD_SET(clientSockFd, &writefds);
	FD_SET(clientSockFd, &readfds);
	select(clientSockFd+1,&readfds,&writefds,NULL, &timeoutDelay);
	
	if (FD_ISSET(clientSockFd, &writefds)
	    &&
	    FD_ISSET(clientSockFd, &readfds)) return READY_TO_READ_AND_WRITE;
	if (FD_ISSET(clientSockFd, &writefds)) return READY_TO_WRITE;
	if (FD_ISSET(clientSockFd, &readfds)) return READY_TO_READ;

	return NOT_READY;
}

int CAVERNnet_tcpClient_c::read(char *ptr, int *nbytes, int blockingType)
{

    //used for bandwidth calculations
 //   int totalDataRead =0;    

	struct timeval timeoutDelay;
	fd_set readfds;

	if (clientSockFd == 0) {
		printf("CAVERNnet_tcpClient_c::Read: This connection is not open.\n");
		return SOCKET_NOT_OPEN;
	}

	int nleft, nread;
	nleft = *nbytes;
	timeOutStatus = 0;

	while (nleft > 0) {

        /*The totalDataRead is incremented  each time something is read - in case of an error this is passed on to the performance monitoring instance using the incrementDataRead() function for bandwidth calculations - to calculate the bandwidth for the data read till the break in loop*/

        FD_ZERO(&readfds);
		FD_SET(clientSockFd, &readfds);

		if (blockingType == NON_BLOCKING) {
			timeoutDelay.tv_sec = 0;
			timeoutDelay.tv_usec = 0;
		}
		if (blockingType == BLOCKING) {
			timeoutDelay.tv_sec = timeOutPeriod;
			timeoutDelay.tv_usec = 0;
		}

		errno = 0;

		// Wait for some file descriptor to be ready
		// If no timeout then wait indefinitely until data arrives
#ifdef WIN32
		int error;

		if ((blockingType == BLOCKING) &&
			(timeOutPeriod == NO_TIME_OUT))
				error = select(clientSockFd+1,&readfds,NULL,NULL, NULL);
		else
			error = select(clientSockFd+1,&readfds,NULL,NULL, &timeoutDelay);
		if (error == SOCKET_ERROR) {
			errno = WSAGetLastError();
			return errno;
		}
#else
		// This one may not work as it should be..
		// Unix socket's man page doesn't say select will alter the
		// value of errno...
		if ((blockingType == BLOCKING) &&
			(timeOutPeriod == NO_TIME_OUT))
				select(clientSockFd+1,&readfds,NULL,NULL, NULL);
		else
			select(clientSockFd+1,&readfds,NULL,NULL, &timeoutDelay);
		if (errno) {
			return errno;
		}
#endif


		// If fd is not set
		if (!(FD_ISSET(clientSockFd, &readfds))) {
			// If non blocking then it simply means there is no data.
			if (blockingType == NON_BLOCKING) 
				return NON_BLOCKING_HAS_NO_DATA;

			// If blocking then timed out waiting for data.
			if (blockingType == BLOCKING) {
				timeOutStatus = 1;
				return TIMED_OUT;
			}

		}

		// If it passes this stage then things are set as blocking.
		blockingType = BLOCKING;

#ifdef WIN32
		nread = ::recv(clientSockFd, ptr, nleft, 0);

		if (nread == SOCKET_ERROR) {
			*nbytes = nread;
			errno = WSAGetLastError();
			if (errno == WSAECONNRESET) 
				return CONNECTION_TERMINATED;
			return errno;
		}
#else
		errno = 0;
		nread = ::read(clientSockFd,ptr,nleft);

		if (errno) {
			*nbytes = nread;
			return errno;
		}
#endif

		if (nread == 0) {
			*nbytes = (*nbytes - nleft);
			return CONNECTION_TERMINATED;
		}

		if (nread > 0) {
			nleft -= nread;
			ptr += nread;
            pmonitor.incrementDataRead(nread);
		}

	}
	*nbytes = *nbytes - nleft;
    return OK;
}

int CAVERNnet_tcpClient_c::write(char *ptr, int  *nbytes, int blockingType)
{
    //used for bandwidth calculations
//    int totalDataSent =0;

	// Place here to prevent crashes when write occurs to a socket
	// closed by an external source.
#ifndef WIN32
	signal(SIGPIPE,SIG_IGN);
#endif

	struct timeval timeoutDelay;
	fd_set writefds;

	if (clientSockFd == 0) {
		printf("CAVERNnet_tcpClient_c::Write: This connection is not open.\n");
		return SOCKET_NOT_OPEN;
	}

	int nleft, nwrite;
	nleft = *nbytes;
	timeOutStatus = 0;

	while (nleft > 0) {
		
  /*The totalDataSent is incremented  each time something is written - in case of an error this is passed on to the performance monitoring instance using the incrementDataSent() function for bandwidth calculations - to calculate the bandwidth for the data written till the break in the loop*/

        FD_ZERO(&writefds);
		FD_SET(clientSockFd, &writefds);

		if (blockingType == NON_BLOCKING) {
			timeoutDelay.tv_sec = 0;
			timeoutDelay.tv_usec = 0;
		}
		if (blockingType == BLOCKING) {
			timeoutDelay.tv_sec = timeOutPeriod;
			timeoutDelay.tv_usec = 0;
		}

		errno = 0;

		// Wait for some file descriptor to be ready
#ifdef WIN32
		int error;

		if ((blockingType == BLOCKING) &&
			(timeOutPeriod == NO_TIME_OUT))
			error = select(clientSockFd+1,NULL,&writefds,NULL, NULL);
		else
			error = select(clientSockFd+1,NULL,&writefds,NULL, &timeoutDelay);
		if (error == SOCKET_ERROR) {
			errno = WSAGetLastError();
			return errno;
		}
#else
		// This one may not work as it should be..
		// Unix socket's man page doesn't say select will alter the
		// value of errno...
		if ((blockingType == BLOCKING) &&
			(timeOutPeriod == NO_TIME_OUT))
			select(clientSockFd+1,NULL,&writefds,NULL, NULL);
		else
			select(clientSockFd+1,NULL,&writefds,NULL, &timeoutDelay);
		if (errno) {
			return errno;
		}
#endif

		// If fd is not set
		if (!(FD_ISSET(clientSockFd, &writefds))) {

			// If non blocking then it simply means there is no data.
			if (blockingType == NON_BLOCKING) 
				return NON_BLOCKING_NOT_READY_TO_WRITE;

			// If blocking then timed out waiting for data.
			if (blockingType == BLOCKING) {
				timeOutStatus = 1;
				return TIMED_OUT;
			}
		}


		// If it passes this stage then things are set as blocking.
		blockingType = BLOCKING;

#ifdef WIN32
		nwrite = ::send(clientSockFd, ptr, nleft, 0);
        if (nwrite == SOCKET_ERROR) {
		    *nbytes = nwrite;
            errno = WSAGetLastError(); 
		    return errno;
		}
#else
		errno = 0;
		nwrite = ::write(clientSockFd,ptr,nleft);

		if (errno) {
			*nbytes = nwrite;
			return errno;
		}
#endif

		if (nwrite == 0) {
			*nbytes = (*nbytes - nleft);
			return CONNECTION_TERMINATED;
		}

		if (nwrite > 0) {
			nleft -= nwrite;
			ptr += nwrite;
            pmonitor.incrementDataSent(nwrite);
		}

	}
	*nbytes = *nbytes - nleft;
	return OK;

}

int  CAVERNnet_tcpClient_c::connectToServer(char *ip, int port)
{

	if (clientSockFd) {
		printf("CAVERNnet_tcpClient_c::Connect: You must close this connection before opening another.\n");
		return 0;
	}

	char serverAddr[128];

	int status = hostnameToIP(ip, serverAddr);
	

	if (status == 0) {
		printf("CAVERNnet_tcpClient_c::Connect: Host IP address %s is invalid.\n", ip);
		return 0;
	}

	/* Fill in the structure "clientInfo" with the address of the
	 * server that we want to connect with.
	 */



	memset((char *) &clientInfo,0, sizeof(clientInfo));

	clientInfo.sin_family = AF_INET;
	clientInfo.sin_addr.s_addr = inet_addr(serverAddr);
	clientInfo.sin_port = htons(port);


	/* Open a TCP socket (an Internet stream socket). */
#ifdef WIN32
	if ((clientSockFd = socket(AF_INET,SOCK_STREAM,0)) == INVALID_SOCKET) {
		//		printf("CAVERNnet_tcpClient_c::Connect: Can't open stream socket.\n");
		return 0;
	}
#else
	if ((clientSockFd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
		//		printf("CAVERNnet_tcpClient_c::Connect: Can't open stream socket.\n");
		return 0;
	}
#endif

	int optlen;
	int sendbuff;

	/* Make sure the buffer is big enough */
	sendbuff = DEFAULT_BUFFER_SIZE;

	optlen = sizeof(sendbuff);
	if (setsockopt(clientSockFd,SOL_SOCKET, SO_SNDBUF, (char *) &sendbuff, optlen) < 0)
		printf("CAVERNnet_tcpClient_c::init: Cannot create a large enough socket buffer.\n");

	/* Set no delay to stop TCP/IP from buffering */
	int noDelay = 1;
	optlen = sizeof(noDelay);

	if (setsockopt(clientSockFd,IPPROTO_TCP,TCP_NODELAY, (char *) &noDelay, optlen) < 0) 
		printf("CAVERNnet_tcpClient_c::Open: Cannot set no delay, you may experience sluggish performance.\n");

	/* Use this to try to eliminate the problem of being unable to rebind to a static port */
	int reuseFlag = 1;
	optlen = sizeof(reuseFlag);
	if (setsockopt(clientSockFd,SOL_SOCKET,SO_REUSEADDR, (char *) &reuseFlag, optlen) < 0) 
		printf("CAVERNnet_tcpClient_c::Open: Cannot reuse port address.\n");
	/***** Bind our local address so that the client can send to us. */

	struct linger lingerData;
	lingerData.l_onoff = 1;
	lingerData.l_linger = 1000;

	optlen = sizeof(lingerData);
	if (setsockopt(clientSockFd,SOL_SOCKET,SO_LINGER, (char *) &lingerData, optlen) < 0)
		printf("CAVERNnet_tcpClient_c::Open: setsockopt: SO_LINGER failed.\n");

	/* Connect to server. */
    int connectResult;
	if ((connectResult = connect(clientSockFd, (struct sockaddr *) &clientInfo, sizeof(clientInfo))) < 0) {
		printf("CAVERNnet_tcpClient_c::Connect: Can't connect to server.\n");
		return connectResult;
	}
	return clientSockFd;
}

void CAVERNnet_tcpClient_c::makeNonBlocking()
{
#ifdef __sgi
	fcntl(clientSockFd, F_SETFL, FNDELAY);
#endif
#ifdef linux
	fcntl(clientSockFd, F_SETFL, O_NONBLOCK);
#endif
#ifdef sun
	fcntl(clientSockFd, F_SETFL, O_NDELAY);
#endif
#ifdef WIN32
    u_long nonBlockingMode = 1L;

    ioctlsocket(clientSockFd, FIONBIO, &nonBlockingMode);
#endif
}

void CAVERNnet_tcpClient_c::close() {
#ifdef WIN32
	if (clientSockFd > 0) ::closesocket(clientSockFd);
#else
	if (clientSockFd > 0) ::close(clientSockFd);
#endif
	clientSockFd = 0;
}

CAVERNnet_tcpClient_c::CAVERNnet_tcpClient_c() {
	timeOutPeriod = CAVERNnet_tcpClient_c::DEFAULT_TIME_OUT;
	clientSockFd = 0;

}

void CAVERNnet_tcpClient_c::showStatus(int status, int nbytes)
{
	switch (status) {
	case SOCKET_NOT_OPEN:
		printf("CAVERNnet_tcpClient_c::showStatus: Socket not open.\n");
		break;
	case CONNECTION_TERMINATED:
		printf("CAVERNnet_tcpClient_c::showStatus: Connection terminated with %d bytes transacted.\n", nbytes);
		break;
	case NON_BLOCKING_HAS_NO_DATA:
		printf("CAVERNnet_tcpClient_c::showStatus: Non-blocking read found no data on stream.\n");
		break;
	case NON_BLOCKING_NOT_READY_TO_WRITE:
		printf("CAVERNnet_tcpClient_c::showStatus: Non-blocking connection is not ready to write.\n");
		break;
	case TIMED_OUT:
		printf("CAVERNnet_tcpClient_c::showStatus: Transaction timed out after %d seconds.\n", timeOutPeriod);
		break;
		//	case NON_BLOCKING_WROTE_NO_DATA:
		//		printf("Non-blocking write wrote no data.\n");
		//		break;
	case OK:
		printf("CAVERNnet_tcpClient_c::showStatus: Transaction ok.\n");
		break;
	default:
		printf("CAVERNnet_tcpClient_c::showStatus: UNIX Socket error no: %d ",errno);
		perror("");
		break;
	}
	fflush(stdout);
}


unsigned int CAVERNnet_tcpClient_c :: getRemoteIP() {
	return clientInfo.sin_addr.s_addr;
}

void CAVERNnet_tcpClient_c :: getRemoteIP(char* name)
{
	ipNumToString(getRemoteIP(),name);

}

unsigned int CAVERNnet_tcpClient_c :: getSelfIP(){
	struct sockaddr_in name;
	int size = sizeof(name);
	getsockname (clientSockFd, (sockaddr*)&name, &size);
	return name.sin_addr.s_addr;
}

void CAVERNnet_tcpClient_c :: getSelfIP(char *name)
{
	ipNumToString(getSelfIP(),name);
}

int CAVERNnet_tcpClient_c :: getSelfPort() {
	struct sockaddr_in name;
	int size = sizeof(name);
	getsockname (clientSockFd, (sockaddr *)&name, &size);
	return ntohs(name.sin_port);
}

/* Get the port # of the client connection */
int CAVERNnet_tcpClient_c :: getRemotePort() {
	return ntohs(clientInfo.sin_port);
}

//Functions added for performance monitoring interface

void CAVERNnet_tcpClient_c ::showStats(char* streamInfo, char* comment)        
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
    getSelfIP(selfIp);
    getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(getSelfPort(), getRemotePort());
    pmonitor.showStats(streamInfo, comment);
    
}

int CAVERNnet_tcpClient_c ::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
    getSelfIP(selfIp);
    getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(getSelfPort(), getRemotePort());
    return pmonitor.logStats(streamInfo, comment, filePtr);
}


int CAVERNnet_tcpClient_c::sendStats(char* streamInfo, char* comment)
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
    getSelfIP(selfIp);
    getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(getSelfPort(), getRemotePort());
    return pmonitor.sendStats(streamInfo, comment);
    
}

int CAVERNnet_tcpClient_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void CAVERNnet_tcpClient_c::exitSendStats()
{
    pmonitor.exitSendStats();
}

void CAVERNnet_tcpClient_c::setInstantaneousLatency(double latency)
{
    pmonitor.setInstantLatency(latency);    
}




