/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERNnet_udpReflector_c.hxx"

const int CAVERNnet_udpReflector_c::FAILED = 0;
const int CAVERNnet_udpReflector_c::OK = 1;
const int CAVERNnet_udpReflector_c::SKIP_DISTRIBUTION = 2;

CAVERNnet_udpReflector_c :: CAVERNnet_udpReflector_c()
{
	itsInit = 0;
	itsTimeOutTime = CAVERNNET_UDP_REFLECTOR_DEFAULT_TIMEOUT;
	itsNumberOfClients = 0;
	itsMainUDPSocket = NULL;
	doMcast = 0;
	doForcedUDP = 0;
	itsLastTime = 0;
	itsUDPPort = 0;
	itsMcastTTL = CAVERNNET_UDP_REFLECTOR_MCAST_TTL;
    	inBuffer = new char[CAVERNNET_UDP_REFLECTOR_MAX_MSG_SIZE];
	interceptCallback = NULL;
	interceptUserData = NULL;
	perfDaemonIP = NULL;
    	perfDaemonPort = PERF_DAEMON_DEFAULT_PORT;
    	perfDaemonStatus = PERFDAEMON_NOT_ALIVE ;
}


void CAVERNnet_udpReflector_c :: setMulticastTTL(int ttl)
{
	itsMcastTTL = ttl;
}

void CAVERNnet_udpReflector_c :: setIncomingPort(unsigned short port)
{
	itsUDPPort = port;
}

void CAVERNnet_udpReflector_c :: setTimeOutTime(int secs)
{
	itsTimeOutTime = secs;
}

void CAVERNnet_udpReflector_c :: setForcedDestination(char *ip, unsigned short port)
{
	itsForcedUDPPort = port;
	strcpy(itsForcedUDPIP,ip);
	doForcedUDP = 1;
}

void CAVERNnet_udpReflector_c:: setMulticastAddress(char *ip, unsigned short port)
{
	doMcast  = 1;
	itsMcastPort = port;
	strcpy(itsMcastIP, ip);
}

// Destructor
CAVERNnet_udpReflector_c :: ~CAVERNnet_udpReflector_c()
{
	if (perfDaemonIP)
       {
            exitSendStats();
            delete perfDaemonIP;
       }
    if (itsMainUDPSocket) delete itsMainUDPSocket;
	CAVERNnet_udp_c *aClient;

	for (int i = 0; i < itsNumberOfClients; i++) {
		aClient = itsBunchOfClients[i];
		if (aClient) delete aClient;
	}
	if (itsMcastSocket) delete itsMcastSocket;
}

int CAVERNnet_udpReflector_c :: init()
{

	if (itsInit) {
		printf("CAVERNnet_udpReflector_c :: init: Initialization already done before\n");
		return CAVERNnet_udpReflector_c::FAILED;
	}

	itsInit = 1;


	itsMainUDPSocket = new CAVERNnet_udp_c;

	if (!itsMainUDPSocket) {
		printf("CAVERNnet_udpReflector_c :: init: UDP cannot allocate memory for socket.\n");
		return CAVERNnet_udpReflector_c::FAILED;
	}

	if (itsMainUDPSocket->init(itsUDPPort) == 0) {
		printf("CAVERNnet_udpReflector_c :: init: UDP Init failed.\n");
		return CAVERNnet_udpReflector_c::FAILED;
	}

	itsMainUDPSocket->makeNonBlocking();

	if (initClientList() == 0) {
		printf("CAVERNnet_udpReflector_c :: init: unable to allocate memory.\n");
		return CAVERNnet_udpReflector_c::FAILED;
	}


	if (doForcedUDP) {
		itsForcedUDP = itsMainUDPSocket->clone();
		itsForcedUDP->setSendAddress(itsForcedUDPIP, itsForcedUDPPort);
		addClient(itsForcedUDP);
	}

	itsMcastSocket = new CAVERNnet_mcast_c;

	if (!itsMcastSocket) {
		printf("CAVERNnet_udpReflector_c :: init: unable to allocate memory for multicast.\n");
		return CAVERNnet_udpReflector_c::FAILED;
	}

	if (doMcast) {
		itsMcastSocket->init(itsMcastIP, itsMcastPort, itsMcastTTL);
		itsMcastSocket->makeNonBlocking();
	}
	return CAVERNnet_udpReflector_c::OK;
}

int CAVERNnet_udpReflector_c :: addClient(CAVERNnet_udp_c *udp)
{
	if (itsNumberOfClients >= CAVERNNET_UDP_REFLECTOR_MAX_NUM_CLIENTS)
		return 0;

	itsBunchOfClients[itsNumberOfClients] = udp;
	/*
    If perfdaemon is alive and initSendStats has been done already by the other clients, then this is a new client which has connected to the reflector and a initSendStats has to be done for this client too
     */
    if (perfDaemonIP)
    {
        if (itsBunchOfClients[itsNumberOfClients]->initSendStats(perfDaemonIP, perfDaemonPort) != CAVERNnet_perfMonitor_c::OK)
            perfDaemonStatus = PERFDAEMON_NOT_ALIVE;               
    }
    itsAliveClients[itsNumberOfClients] = getTimeInSecs();
	itsNumberOfClients++;
	return 1;
}

int CAVERNnet_udpReflector_c :: initClientList()
{
	itsNumberOfClients = 0;

	itsAliveClients = new double[CAVERNNET_UDP_REFLECTOR_MAX_NUM_CLIENTS];
	if (!itsAliveClients) return 0;

	itsBunchOfClients = new CAVERNnet_udp_c*[CAVERNNET_UDP_REFLECTOR_MAX_NUM_CLIENTS];
	if (!itsBunchOfClients) return 0;

	for (int i = 0; i < CAVERNNET_UDP_REFLECTOR_MAX_NUM_CLIENTS; i++){
		itsAliveClients[i] = 0;
		itsBunchOfClients[i] = NULL;
	}
	return 1;
}

void CAVERNnet_udpReflector_c:: intercept(int (*callback)(CAVERNnet_udp_c*, char** buffer, int* datasize, void* userData), void* userData)
{
    interceptCallback = callback;
    interceptUserData = userData;
}

int CAVERNnet_udpReflector_c :: process()
{
	int currentConnectionIndex;
	int anError = 1;
    int distribution = CAVERNnet_udpReflector_c::OK;
    
	if (doMcast) {
		int realSize = itsMcastSocket->receive(inBuffer, CAVERNNET_UDP_REFLECTOR_MAX_MSG_SIZE);
		if (realSize > 0) {
			incrementDataRead(realSize);
			replyToClients(inBuffer, realSize, CAVERNNET_UDP_REFLECTOR_SEND_TO_ALL_UDP_CLIENTS);
		}
	}

	int realSize = itsMainUDPSocket->receive(inBuffer, CAVERNNET_UDP_REFLECTOR_MAX_MSG_SIZE);
	if (realSize > 0) {
		incrementDataRead(realSize);

		currentConnectionIndex = checkClient(itsMainUDPSocket);
		if (currentConnectionIndex >= 0) {
			if (doMcast) {
				itsMcastSocket->send(inBuffer, realSize);
				incrementDataSent(realSize);
			}
            if (interceptCallback)
                distribution = interceptCallback(itsBunchOfClients[currentConnectionIndex], &inBuffer,&realSize, interceptUserData);
            if (distribution != CAVERNnet_udpReflector_c::SKIP_DISTRIBUTION)
           	    replyToClients(inBuffer, realSize, currentConnectionIndex);
		} else {
			printf("CAVERNnet_udpReflector_c :: process: Can't handle any more clients\n");
			anError = 0;
		}

	} else {
#ifdef WIN32
		if (WSAGetLastError() != WSAEWOULDBLOCK && WSAGetLastError() != 0) {
			printf("CAVERNnet_udpReflector_c :: process: UDP RECEIVE ERROR %d\n", WSAGetLastError());
			anError = 0;
		}
#else
		if (errno != EWOULDBLOCK && errno!= 0) {
			printf("CAVERNnet_udpReflector_c :: process: UDP RECEIVE ERROR %d\n", errno);
			anError = 0;
		}
#endif
	}

	purgeOldClients();

	return anError;

}

int CAVERNnet_udpReflector_c :: checkClient(CAVERNnet_udp_c *mainUDPSocket)
{

	CAVERNnet_udp_c *eachClient;

	for (int i = 0; i < itsNumberOfClients; i++) {
		eachClient = itsBunchOfClients[i];
		if ((eachClient->getSendIP() == mainUDPSocket->getReceiveIP()) &&
		    (eachClient->getSendPort() == mainUDPSocket->getReceivePort())) {
			itsAliveClients[i] = getTimeInSecs();
			return i;
		}
	}
	eachClient = mainUDPSocket->clone();
	eachClient->copyReceiveAddressToSendAddress();
	if (addClient(eachClient) == 0) {
		return -1;
	}

	printf("CAVERNnet_udpReflector_c: New Client: %d\n",itsNumberOfClients-1);
	eachClient->printInfo();
	return itsNumberOfClients-1;
}

void CAVERNnet_udpReflector_c :: purgeOldClients()
{
	double clock = getTimeInSecs();

	if ((clock - itsLastTime) < double(itsTimeOutTime)) return;

	// printf("Purging...\n"); fflush(stdout);
	itsLastTime = clock;

	int start = 0;

	// If forced udp skip past the 0th client.
	if (doForcedUDP) start = 1;

	for (int i = start ; i< itsNumberOfClients; i++) {
		if ((clock - itsAliveClients[i]) > double(itsTimeOutTime)) {
			printf("Purging client:\n");
			itsBunchOfClients[i]->printInfo();

			delete itsBunchOfClients[i];
			
			itsBunchOfClients[i] = itsBunchOfClients[itsNumberOfClients-1];
			itsAliveClients[i] = itsAliveClients[itsNumberOfClients-1];
			itsAliveClients[itsNumberOfClients-1] = 0;
			itsNumberOfClients--;
		}
	}
}

/* Reply to all clients that are conencted to me. */
void CAVERNnet_udpReflector_c :: replyToClients(char *mesg, int size, int exceptClientIndex)
{
	for(int i = 0 ; i<itsNumberOfClients; i++) {
		if (i != exceptClientIndex) {
			itsBunchOfClients[i]->send(mesg, size);
			incrementDataSent(size);
		}
	}
}

void CAVERNnet_udpReflector_c ::enableInstrumentation()
{
    if (itsMainUDPSocket)
        itsMainUDPSocket->enableInstrumentation();
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
               itsBunchOfClients[i]->enableInstrumentation();        
        }
}

void CAVERNnet_udpReflector_c ::disableInstrumentation()
{
    if(itsMainUDPSocket)
        itsMainUDPSocket->disableInstrumentation();
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                itsBunchOfClients[i]->disableInstrumentation();
        }
}


void CAVERNnet_udpReflector_c ::showStats(char* streamInfo, char* comment)
{
    if(itsMainUDPSocket)
        itsMainUDPSocket->showStats(streamInfo, comment);
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                itsBunchOfClients[i]->showStats(streamInfo, comment);
        }
}

int CAVERNnet_udpReflector_c ::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    int status=1;
    if(itsMainUDPSocket)    
        status = itsMainUDPSocket->logStats(streamInfo, comment, filePtr);
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                status = ((itsBunchOfClients[i]->logStats(streamInfo, comment, filePtr))&& status);
        }
     if (status)    
        return CAVERNnet_perfMonitor_c::OK;
     else return CAVERNnet_perfMonitor_c::FAILED;                              
}

int CAVERNnet_udpReflector_c ::sendStats(char* streamInfo, char* comment)
{
    int status=1;
    if(itsMainUDPSocket)
        status = itsMainUDPSocket->sendStats(streamInfo, comment);
    for(int i=0; i<itsNumberOfClients; i++)
       {
           if(itsBunchOfClients[i])               
           status = ((itsBunchOfClients[i]->sendStats(streamInfo, comment))&& status);
       }
    if (status)          
      return CAVERNnet_perfMonitor_c::OK;
    else return CAVERNnet_perfMonitor_c::FAILED;                            
}

int CAVERNnet_udpReflector_c ::initSendStats(char* monitorClientIP, int port)
{
    int status=1;
    if (monitorClientIP == NULL) return CAVERNnet_perfMonitor_c::FAILED;
    
    perfDaemonIP = new char[strlen(monitorClientIP) + 1];
    strcpy(perfDaemonIP, monitorClientIP);
    perfDaemonPort = port;
            
    if(itsMainUDPSocket)
        status = itsMainUDPSocket->initSendStats(monitorClientIP, port);
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                status = ((itsBunchOfClients[i]->initSendStats(monitorClientIP, port))&&status);
        }
    if (status)
    {
        perfDaemonStatus = PERFDAEMON_ALIVE;
        return CAVERNnet_perfMonitor_c::OK;
    }    
    else
    {
        perfDaemonStatus = PERFDAEMON_NOT_ALIVE;
        return CAVERNnet_perfMonitor_c::FAILED; 
    }    
}

void CAVERNnet_udpReflector_c ::exitSendStats()
{
    if (itsMainUDPSocket)
        itsMainUDPSocket->exitSendStats();
    for (int i=0; i<itsNumberOfClients; i++)
    {
        if(itsBunchOfClients[i])
            itsBunchOfClients[i]->exitSendStats();
    }
}





























