/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/
   
#include "CAVERNnet_udp_c.hxx"
#include "perfmon.hxx"

const int CAVERNnet_udp_c::OK = 1;
const int CAVERNnet_udp_c::FAILED = 0;
const int CAVERNnet_udp_c::BLOCKING = 1;
const int CAVERNnet_udp_c::NON_BLOCKING = 2;
const int CAVERNnet_udp_c::NULL_VALUE = 3;
const int CAVERNnet_udp_c::NO_TIME_OUT = -1;
const int CAVERNnet_udp_c::DEFAULT_TIME_OUT = NO_TIME_OUT; 
const int CAVERNnet_udp_c::SOCKET_NOT_OPEN = -1;
const int CAVERNnet_udp_c::TIMED_OUT = -4;
const int CAVERNnet_udp_c::NON_BLOCKING_NOT_READY_TO_WRITE = -5;
const int CAVERNnet_udp_c::NON_BLOCKING_HAS_NO_DATA = -6;

#ifdef sun
long gethostid(void);
#endif

#ifdef WIN32
#define MAXHOSTNAME   128   /* Maximum length of a DNS hostname */

// This function is excerpted from "Windows Sockets Network Programming"
// by Bob Quinn and Dave Shute published from Addison-Wesley.
static long gethostid(void)
{
	char szLclHost [MAXHOSTNAME];
	LPHOSTENT lpstHostent;
	SOCKADDR_IN stLclAddr;
	SOCKADDR_IN stRmtAddr;
	int nAddrSize = sizeof(SOCKADDR);
	SOCKET hSock;
	int nRet;

	/* Init local address (to zero) */
	stLclAddr.sin_addr.s_addr = INADDR_ANY;

	/* Get the local hostname */
	nRet = gethostname(szLclHost, MAXHOSTNAME); 
	if (nRet != SOCKET_ERROR) {
		/* Resolve hostname for local address */
		lpstHostent = gethostbyname((LPSTR)szLclHost);
		if (lpstHostent)
			stLclAddr.sin_addr.s_addr = *((u_long FAR*) (lpstHostent->h_addr));
	} 

	/* If still not resolved, then try second strategy */
	if (stLclAddr.sin_addr.s_addr == INADDR_ANY) {
		/* Get a UDP socket */
		hSock = socket(AF_INET, SOCK_DGRAM, 0);
		if (hSock != INVALID_SOCKET)  {
			/* Connect to arbitrary port and address (NOT loopback) */
			stRmtAddr.sin_family = AF_INET;
			stRmtAddr.sin_port   = htons(IPPORT_ECHO);
			stRmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
			nRet = connect(hSock, (LPSOCKADDR)&stRmtAddr,
						   sizeof(SOCKADDR));
			if (nRet != SOCKET_ERROR) {
				/* Get local address */
				getsockname(hSock, (LPSOCKADDR)&stLclAddr, 
							(int FAR*)&nAddrSize);
			}
			closesocket(hSock);   /* we're done with the socket */
		}
	}
	return (stLclAddr.sin_addr.s_addr);
}
#endif

CAVERNnet_udp_c :: CAVERNnet_udp_c ()
{
	typeOfBlocking = CAVERNnet_udp_c::BLOCKING;
	timeOutPeriod = CAVERNnet_udp_c::DEFAULT_TIME_OUT;
	enableInstr = 0;
}

void CAVERNnet_udp_c :: makeNonBlocking()
{
#ifdef __sgi
	fcntl(sockfd, F_SETFL, FNDELAY);
#endif
#ifdef linux
	fcntl(sockfd, F_SETFL, O_NONBLOCK);
#endif
#ifdef sun
	fcntl(sockfd, F_SETFL, O_NDELAY);
#endif
#ifdef WIN32
    u_long nonBlockingMode = 1L;

    ioctlsocket(sockfd, FIONBIO, &nonBlockingMode);
#endif

typeOfBlocking = CAVERNnet_udp_c::NON_BLOCKING;
}

int CAVERNnet_udp_c :: init(unsigned short port)
{
	/* Open datagram socket */
#ifdef WIN32
	if ((sockfd = socket(AF_INET, SOCK_DGRAM,0)) == INVALID_SOCKET) {
		printf("CAVERNnet_udp_c::init: cannot open datagram socket.\n");
		return FAILED;
	}
#else
	if ((sockfd = socket(AF_INET, SOCK_DGRAM,0)) < 0) {
		printf("CAVERNnet_udp_c::init: cannot open datagram socket.\n");
		return FAILED;
	}
#endif

	memset((char *) &selfAddress,0, sizeof(selfAddress));
	selfAddress.sin_family = AF_INET;
	selfAddress.sin_addr.s_addr = htonl(INADDR_ANY);
	selfAddress.sin_port = htons(port);

	if (bind(sockfd, (struct sockaddr *) &selfAddress, sizeof(selfAddress)) < 0) {
		printf("CAVERNnet_udp_c::init: Cannot bind local address.\n");
		return FAILED;
	}

	return OK;
}


int CAVERNnet_udp_c::send(char *message, int length, int blockingType)
{
	char* newmsg = message;
	int newlength = length;

	struct timeval timeoutDelay;
	fd_set writefds;

	if (sockfd == 0) {
		printf("CAVERNnet_udp_c::send: This connection is not open.\n");
		return SOCKET_NOT_OPEN;
	}

	//Check the blocking type and set the timeout values

	if (blockingType == NULL_VALUE)
		blockingType = typeOfBlocking;
 
	timeOutStatus = 0;

        FD_ZERO(&writefds);
	FD_SET(sockfd, &writefds);

	if (blockingType == NON_BLOCKING) {
		timeoutDelay.tv_sec = 0;
		timeoutDelay.tv_usec = 0;
	}
	if (blockingType == BLOCKING) {
		timeoutDelay.tv_sec = timeOutPeriod;
		timeoutDelay.tv_usec = 0;
	}

	errno = 0;

	// Wait for some file descriptor to be ready
#ifdef WIN32
	int error;

	if ((blockingType == BLOCKING) &&
		(timeOutPeriod == NO_TIME_OUT))
		error = select(sockfd+1,NULL,&writefds,NULL, NULL);
	else
		error = select(sockfd+1,NULL,&writefds,NULL, &timeoutDelay);
	if (error == SOCKET_ERROR) {
		errno = WSAGetLastError();
		return errno;
	}
#else
	if ((blockingType == BLOCKING) &&
		(timeOutPeriod == NO_TIME_OUT))
			select(sockfd+1,NULL,&writefds,NULL, NULL);
	else
			select(sockfd+1,NULL,&writefds,NULL, &timeoutDelay);
	if (errno) {
		return errno;
	}
#endif

	// If fd is not set
	if (!(FD_ISSET(sockfd, &writefds))) {

		// If blocking then timed out waiting for data.
		if (blockingType == BLOCKING) {
			timeOutStatus = 1;
			return TIMED_OUT;
		}

		// If nonblocking then there is nothing to send.
		if (blockingType == NON_BLOCKING)
			return NON_BLOCKING_NOT_READY_TO_WRITE;
	}

	if (enableInstr) {
		newmsg = _cavern_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;

		_cavern_prependPerformanceMonitoringHeader(newmsg, message, length);
	}

	int n = sendto(sockfd, newmsg, newlength, 0, (sockaddr *)&sendAddress,
		       sizeof(sendAddress));
#ifdef WIN32
	if (n == SOCKET_ERROR) errno = WSAGetLastError();
#endif
	if (n > 0) 
        pmonitor.incrementDataSent(n);
	if (enableInstr) {
		delete newmsg;
		int realsize = n-_cavern_getSizeOfPerformanceMonitoringHeader();
		if (realsize > 0)
			return realsize;
		else
			return 0;
	}
	else return n;
}

int CAVERNnet_udp_c::receive(char *message, int length, int blockingType)
{
	char *newmsg = message;
	int newlength = length;

	struct timeval timeoutDelay;
	fd_set readfds;

	if (sockfd == 0) {
		printf("CAVERNnet_udp_c::receive: This connection is not open.\n");
		return SOCKET_NOT_OPEN;
	}

	//Check the blocking type and set the timeout values
	
	if (blockingType == NULL_VALUE)
		blockingType = typeOfBlocking;
	
    timeOutStatus = 0;

        FD_ZERO(&readfds);
	FD_SET(sockfd, &readfds);

	if (blockingType == NON_BLOCKING) {
		timeoutDelay.tv_sec = 0;
		timeoutDelay.tv_usec = 0;
	}
	if (blockingType == BLOCKING) {
		timeoutDelay.tv_sec = timeOutPeriod;
		timeoutDelay.tv_usec = 0;
	}

	errno = 0;

	// Wait for some file descriptor to be ready
	// If no timeout then wait indefinitely until data arrives
#ifdef WIN32
	int error;

	if ((blockingType == BLOCKING) &&
		(timeOutPeriod == NO_TIME_OUT))
			error = select(sockfd+1,&readfds,NULL,NULL, NULL);
		else
			error = select(sockfd+1,&readfds,NULL,NULL, &timeoutDelay);
		if (error == SOCKET_ERROR) {
			errno = WSAGetLastError();
			return errno;
		}
#else
		if ((blockingType == BLOCKING) &&
			(timeOutPeriod == NO_TIME_OUT))
				select(sockfd+1,&readfds,NULL,NULL, NULL);
		else
			select(sockfd+1,&readfds,NULL,NULL, &timeoutDelay);
		if (errno) {
			return errno;
		}
#endif

	// If fd is not set
	if (!(FD_ISSET(sockfd, &readfds))) {

    		// If blocking then timed out waiting for data.
	    	if (blockingType == BLOCKING) {
		       	timeOutStatus = 1;
	    		return TIMED_OUT;
		}
		if (blockingType == NON_BLOCKING) 
			return NON_BLOCKING_HAS_NO_DATA;
	}


	// if instrumentation enabled
	if (enableInstr) {
		newmsg = _cavern_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;
	}


	int clilen = sizeof(receiveAddress);
	int n;

	n = recvfrom(sockfd, newmsg, newlength, 0, (struct sockaddr *) &receiveAddress,& clilen);
	if (n > 0) {
        pmonitor.incrementDataRead(n);

		// if instrumentation enabled
		if (enableInstr) {
			double remTimeStamp, latency;
			int realSize;
			_cavern_unprependPerformanceMonitoringHeader(&remTimeStamp, &latency, newmsg, n, message, &realSize);

			delete newmsg;

			pmonitor.setInstantLatency(latency);
			return realSize;
		}
	} else {
#ifdef WIN32
		if (n == SOCKET_ERROR) errno = WSAGetLastError();
#endif
		if (enableInstr) delete newmsg;
	}

	return n;
}


int CAVERNnet_udp_c::peek(char *message, int length)
{
	int clilen = sizeof(receiveAddress);
	int n;

	char *newmsg = message;
	int newlength = length;

	// if instrumentation enabled
	if (enableInstr) {
		newmsg = _cavern_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;
	}

	n = recvfrom(sockfd, newmsg, newlength, MSG_PEEK, (struct sockaddr *) &receiveAddress, & clilen);

	// if instrumentation enabled
	if (n>0) {
        pmonitor.incrementDataRead(n);
		if (enableInstr) {
			double remoteTimeStamp, latency;
			int realSize;

			_cavern_unprependPerformanceMonitoringHeader(&remoteTimeStamp, &latency, newmsg, n, message, &realSize);
			delete newmsg;
            pmonitor.setInstantLatency(latency);
			return realSize;
		}
	} else {
		if (enableInstr) delete newmsg;
	}

	return n;
}

int CAVERNnet_udp_c :: setSendAddress(char *host, unsigned short port)
{

	char destHostIP[32];
	int status;

	status = hostnameToIP(host, destHostIP);

	if (status == 0) return 0;

	memset((char *) &sendAddress,0, sizeof(sendAddress));

	sendAddress.sin_family = AF_INET;
	sendAddress.sin_addr.s_addr = inet_addr(destHostIP);
	sendAddress.sin_port = htons(port);

	return 1;
}

void CAVERNnet_udp_c::copyReceiveAddressToSendAddress()
{
	// Must copy entire structure not just ip and port.
	sendAddress = receiveAddress;
}

void CAVERNnet_udp_c::enableInstrumentation()
{
	enableInstr = 1;
}

void CAVERNnet_udp_c::disableInstrumentation()
{
	enableInstr = 0;
}
void CAVERNnet_udp_c::printInfo() {
	char ip[256];
	getSelfIP(ip);
	printf("CAVERNnet_udp_c: Self IP accepting connections on: %s Port: %u\n", ip, getSelfPort());
	getSendIP(ip);
	printf("CAVERNnet_udp_c: Send IP: %s Port: %u\n", ip, getSendPort());
	getReceiveIP(ip);
	printf("CAVERNnet_udp_c: Receive IP: %s Port: %u\n", ip, getReceivePort());
}

CAVERNnet_udp_c * CAVERNnet_udp_c :: clone()
{
	CAVERNnet_udp_c *duplicate;
	duplicate = new CAVERNnet_udp_c;
	*duplicate = *this;
	return duplicate;
}

unsigned long CAVERNnet_udp_c :: getSelfIP(){

	/* THIS OLD STUFF RETURNS 0 for some reason
	struct sockaddr_in name;
	int size = sizeof(name);
	getsockname (sockfd, &name, &size);
	return name.sin_addr.s_addr;
	*/
#ifdef linux
        unsigned short a1, a2;
        unsigned long ip, temp;
       
        ip = gethostid();
        a1 = (unsigned short) (ip >> 16);
        a2 = (unsigned short) ip; 
        temp = (unsigned long) a2;
        temp <<= 16;
	temp |= a1;
        return temp;
#else
	return gethostid();
#endif
}

void CAVERNnet_udp_c :: getSelfIP(char *ip) {ipNumToString(getSelfIP(), ip);}

unsigned short  CAVERNnet_udp_c :: getSelfPort() {
	struct sockaddr_in name;
	int size = sizeof(name);
	getsockname (sockfd, (sockaddr *)&name, &size);
	return ntohs(name.sin_port);
}


unsigned long CAVERNnet_udp_c :: getReceiveIP() {
    return receiveAddress.sin_addr.s_addr;
}

void  CAVERNnet_udp_c :: getReceiveIP(char *ip) {
	ipNumToString(receiveAddress.sin_addr.s_addr, ip);
}

unsigned short  CAVERNnet_udp_c :: getReceivePort() {return ntohs(receiveAddress.sin_port);}

unsigned long  CAVERNnet_udp_c :: getSendIP() {return sendAddress.sin_addr.s_addr;}

void  CAVERNnet_udp_c :: getSendIP(char *ip) {
	ipNumToString(sendAddress.sin_addr.s_addr, ip);
}

unsigned short  CAVERNnet_udp_c :: getSendPort()
{
    return ntohs(sendAddress.sin_port);
}

/*For performance monitoring*/

void CAVERNnet_udp_c ::showStats(char* streamInfo, char* comment)        
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
	unsigned short remotePort;

    getSelfIP(selfIp);
	if (getSendIP()) getSendIP(remoteIp);
	else getReceiveIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
	remotePort = (getSendPort()) ? getSendPort() : getReceivePort();
    pmonitor.setPorts(getSelfPort(), remotePort);
    pmonitor.showStats(streamInfo, comment);
    
}

int CAVERNnet_udp_c ::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
	unsigned short remotePort;

    getSelfIP(selfIp);
	if (getSendIP()) getSendIP(remoteIp);
	else getReceiveIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
	remotePort = (getSendPort()) ? getSendPort() : getReceivePort();
    pmonitor.setPorts(getSelfPort(), remotePort);
    return pmonitor.logStats(streamInfo, comment, filePtr);
}

int CAVERNnet_udp_c::sendStats(char* streamInfo, char* comment)
{
    char *selfIp = new char[256];
    char *remoteIp = new char[256];   
	unsigned short remotePort;

    getSelfIP(selfIp);
	if (getSendIP()) getSendIP(remoteIp);
	else getReceiveIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
	remotePort = (getSendPort()) ? getSendPort() : getReceivePort();
    pmonitor.setPorts(getSelfPort(), remotePort);
    return pmonitor.sendStats(streamInfo, comment);
    
}

int CAVERNnet_udp_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void CAVERNnet_udp_c::exitSendStats()
{
    pmonitor.exitSendStats();
}
