/******************************************************************
 * CAVERNsoft
 * Copyright (C) 1994-2000 Electronic Visualization Laboratory,
 * all rights reserved
 * By Jason Leigh, Yong-joo Cho, Naveen Krishnaprasad, Chris Scharver,
 * Stuart Bailey, Atul Nayak, Shalini Venkataraman
 * University of Illinois at Chicago
 * 
 * This publication and its text and code may not be copied for commercial 
 * use without the express written permission of the University of Illinois
 * at Chicago.
 * The contributors disclaim any representation of warranty: use this
 * code at your own risk.
 * Direct questions, comments etc to cavern@evl.uic.edu
 ******************************************************************/

#include "CAVERNts_condition_c.hxx"

CAVERNts_condition_c :: CAVERNts_condition_c()
{
#ifdef CAVERN_USE_GLOBUS_THREADS
	globus_cond_init(&itsCV, NULL);
#endif

#ifdef CAVERN_USE_PTHREADS
	pthread_cond_init(&itsCV, NULL);
#endif

#ifdef CAVERN_USE_WINDOWSTHREADS
    itsCV = new _CAVERN_COND_T;
#endif
}

CAVERNts_condition_c :: ~CAVERNts_condition_c()
{
#ifdef CAVERN_USE_GLOBUS_THREADS
	globus_cond_destroy(&itsCV);
#endif
#ifdef CAVERN_USE_PTHREADS
	pthread_cond_destroy(&itsCV);
#endif
#ifdef CAVERN_USE_WINDOWSTHREADS
  if (itsCV) delete itsCV;
#endif
}

int CAVERNts_condition_c :: wait(CAVERNts_mutex_c *mutex)
{
#ifdef CAVERN_USE_GLOBUS_THREADS
	return globus_cond_wait(&itsCV, mutex->getMutex());
#endif
#ifdef CAVERN_USE_PTHREADS
	return pthread_cond_wait(&itsCV, mutex->getMutex());
#endif
#ifdef CAVERN_USE_WINDOWSTHREADS
  // Prevent race conditions on the <waiters_> count.
  WaitForSingleObject(itsCV->waitersLock, INFINITE);
  itsCV->waiters++;
  ReleaseMutex(itsCV->waitersLock);

  int result = 0;

  // We keep the lock held just long enough to increment the count of
  // waiters by one.  Note that we can't keep it held across the call
  // to ACE_OS::sema_wait() since that will deadlock other calls to
  // ACE_OS::cond_signal().
  mutex->unlock();

  // Wait to be awakened by a ACE_OS::cond_signal() or
  // ACE_OS::cond_broadcast().
  result = WaitForSingleObject(itsCV->sema, INFINITE);
  if (result == WAIT_FAILED) return -1;
  
  // Reacquire lock to avoid race conditions on the <waiters_> count.
  WaitForSingleObject(itsCV->waitersLock, INFINITE);

  // We're ready to return, so there's one less waiter.
  itsCV->waiters--;

  int lastWaiter = itsCV->wasBroadcast && itsCV->waiters == 0;

  // Release the lock so that other collaborating threads can make
  // progress.
  ReleaseMutex(itsCV->waitersLock);

  if (result == -1)
    // Bad things happened, so let's just return below.
    /* NOOP */;
  // If we're the last waiter thread during this particular broadcast
  // then let all the other threads proceed.
  else if (lastWaiter)
    SetEvent(itsCV->waitersDone);

  // We must always regain the <external_mutex>, even when errors
  // occur because that's the guarantee that we give to our callers.
  mutex->lock();

  return result;
#endif
}


int CAVERNts_condition_c :: signal()
{
#ifdef CAVERN_USE_GLOBUS_THREADS
	return globus_cond_signal(&itsCV);
#endif
#ifdef CAVERN_USE_PTHREADS
	return pthread_cond_signal(&itsCV);
#endif
#ifdef CAVERN_USE_WINDOWSTHREADS
    // If there aren't any waiters, then this is a no-op.  Note that
    // this function *must* be called with the <external_mutex> held
    // since other wise there is a race condition that can lead to the
    // lost wakeup bug...  This is needed to ensure that the <waiters_>
    // value is not in an inconsistent internal state while being
    // updated by another thread.
    WaitForSingleObject(itsCV->waitersLock, INFINITE);
    int haveWaiters = itsCV->waiters > 0;
    ReleaseMutex(itsCV->waitersLock);

    if (haveWaiters != 0) return !ReleaseSemaphore(itsCV->sema, 1, NULL);
    else return 0; // No-op
#endif
}

int CAVERNts_condition_c :: broadcastSignal()
{
#ifdef CAVERN_USE_GLOBUS_THREADS
	return globus_cond_broadcast(&itsCV);
#endif
#ifdef CAVERN_USE_PTHREADS
	return pthread_cond_broadcast(&itsCV);
#endif
#ifdef CAVERN_USE_WINDOWSTHREADS
    // The <external_mutex> must be locked before this call is made.

    // This is needed to ensure that <waiters_> and <was_broadcast_> are
    // consistent relative to each other.
    WaitForSingleObject(itsCV->waitersLock, INFINITE);
    int haveWaiters = 0;

    if (itsCV->waiters > 0) {
    // We are broadcasting, even if there is just one waiter...
    // Record the fact that we are broadcasting.  This helps the
    // cond_wait() method know how to optimize itself.  Be sure to
    // set this with the <waiters_lock_> held.
        itsCV->wasBroadcast = 1;
        haveWaiters = 1;
    }
    ReleaseMutex(itsCV->waitersLock);
    int result = 0;
    if (haveWaiters) {
        // Wake up all the waiters.
        if (ReleaseSemaphore(itsCV->sema, 1, NULL) == 0) result = -1;
            // Wait for all the awakened threads to acquire their part of
            // the counting semaphore.
        else {
            switch (WaitForSingleObject(itsCV->waitersDone, INFINITE)) {
                case WAIT_OBJECT_0:
                    result = 0;

                default:
                    result = -1;
            }

            if (result == -1) result = -1;
        }
        
        // This is okay, even without the <waiters_lock_> held because
        // no other waiter threads can wake up to access it.
        itsCV->wasBroadcast = 0;
    }
    return result;
#endif
}

CAVERN_COND_T *CAVERNts_condition_c::getCondition()
{
	return &itsCV;
}
